from wbcore.metadata.configs.endpoints import EndpointViewConfig

from wbcompliance.models import ComplianceType


class ComplianceTypeEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        if ComplianceType.is_administrator(self.request.user):
            return super().get_endpoint()
        return None



from django.contrib import admin
from guardian.admin import GuardedModelAdmin

from wbcompliance.models.risk_management.rules import (
    RiskRule,
    RuleBackend,
    RuleCheckedObjectRelationship,
    RuleGroup,
    RuleThreshold,
)

from .incidents import IncidentInLine


class RuleThresholdInLine(admin.TabularInline):
    model = RuleThreshold
    fields = ["range", "severity", "notifiable_users", "notifiable_groups"]
    extra = 0
    autocomplete_fields = ["notifiable_users"]
    raw_id_fields = ["notifiable_users", "notifiable_groups", "rule"]


class RuleCheckedObjectRelationshipInLine(admin.TabularInline):
    model = RuleCheckedObjectRelationship
    fields = [
        "checked_object_content_type",
        "checked_object_id",
    ]
    extra = 0
    raw_id_fields = ["checked_object_content_type", "rule"]


@admin.register(RuleGroup)
class RuleGroupModelAdmin(admin.ModelAdmin):
    list_display = [
        "key",
        "name",
    ]


@admin.register(RuleBackend)
class RuleBackendModelAdmin(admin.ModelAdmin):
    list_display = [
        "name",
        "rule_group",
        "backend_class_path",
        "backend_class_name",
        "allowed_checked_object_content_type",
    ]
    raw_id_fields = ["allowed_checked_object_content_type"]


@admin.register(RiskRule)
class RiskRuleModelAdmin(GuardedModelAdmin):
    list_display = [
        "name",
        "rule_backend",
        "is_enable",
        "only_passive_check_allowed",
        "is_silent",
        "is_mandatory",
        "parameters",
    ]
    inlines = [RuleThresholdInLine, IncidentInLine, RuleCheckedObjectRelationshipInLine]



from django.utils.translation import gettext as _
from wbcore.menus import ItemPermission, Menu, MenuItem

from wbcompliance.permissions import (
    is_admin_compliance,
    is_internal_employee_compliance,
)

from .compliance_type import COMPLIANCETYPE_MENUITEM

COMPLIANCETASKGROUP_MENUITEM = MenuItem(
    label=_("Groups Indicator"),
    endpoint="wbcompliance:compliancetaskgroup-list",
    permission=ItemPermission(
        method=is_admin_compliance,
        permissions=["wbcompliance.view_compliancetaskgroup", "wbcompliance.administrate_compliance"],
    ),
    add=MenuItem(
        label=_("Add Group Indicator"),
        endpoint="wbcompliance:compliancetaskgroup-list",
        permission=ItemPermission(method=is_admin_compliance, permissions=["wbcompliance.administrate_compliance"]),
    ),
)

COMPLIANCETASK_MENUITEM = MenuItem(
    label=_("Indicators"),
    endpoint="wbcompliance:compliancetask-list",
    permission=ItemPermission(
        method=is_admin_compliance,
        permissions=["wbcompliance.view_compliancetask", "wbcompliance.administrate_compliance"],
    ),
    add=MenuItem(
        label=_("Add Indicator"),
        endpoint="wbcompliance:compliancetask-list",
        permission=ItemPermission(method=is_admin_compliance, permissions=["wbcompliance.administrate_compliance"]),
    ),
)

COMPLIANCETASKINSTANCE_MENUITEM = MenuItem(
    label=_("Instances Indicator"),
    endpoint="wbcompliance:compliancetaskinstance-list",
    permission=ItemPermission(
        method=lambda request: request.user.is_superuser,
        permissions=["wbcompliance.view_compliancetaskinstance", "wbcompliance.administrate_compliance"],
    ),
)

COMPLIANCETASKMATRIX_MENUITEM = MenuItem(
    label=_("Indicator Matrix"),
    endpoint="wbcompliance:compliancetaskmatrix-list",
    permission=ItemPermission(
        method=is_admin_compliance,
        permissions=[
            "wbcompliance.view_compliancetaskinstance",
            "wbcompliance.view_compliancetask",
            "wbcompliance.administrate_compliance",
        ],
    ),
)

COMPLIANCEACTION_MENUITEM = MenuItem(
    label=_("Actions"),
    endpoint="wbcompliance:complianceaction-list",
    permission=ItemPermission(
        method=is_admin_compliance,
        permissions=["wbcompliance.view_complianceaction", "wbcompliance.administrate_compliance"],
    ),
    add=MenuItem(
        label=_("Add Compliance Action"),
        endpoint="wbcompliance:complianceaction-list",
        permission=ItemPermission(method=is_admin_compliance, permissions=["wbcompliance.administrate_compliance"]),
    ),
)

COMPLIANCEEVENT_MENUITEM = MenuItem(
    label=_("Events"),
    endpoint="wbcompliance:complianceevent-list",
    permission=ItemPermission(
        method=is_internal_employee_compliance, permissions=["wbcompliance.view_complianceevent"]
    ),
    add=MenuItem(
        label=_("Add Compliance Event"),
        endpoint="wbcompliance:complianceevent-list",
        permission=ItemPermission(
            method=is_internal_employee_compliance, permissions=["wbcompliance.add_complianceevent"]
        ),
    ),
)

ADMIN_COMPLIANCEEVENT_MENUITEM = MenuItem(
    label=_("Events"),
    endpoint="wbcompliance:complianceevent-list",
    permission=ItemPermission(method=is_admin_compliance, permissions=["wbcompliance.view_complianceevent"]),
    add=MenuItem(
        label=_("Add Compliance Event"),
        endpoint="wbcompliance:complianceevent-list",
        permission=ItemPermission(method=is_admin_compliance, permissions=["wbcompliance.add_complianceevent"]),
    ),
)

REVIEWCOMPLIANCETASK_MENUITEM = MenuItem(
    label=_("Reports"),
    endpoint="wbcompliance:reviewcompliancetask-list",
    permission=ItemPermission(method=is_admin_compliance, permissions=["wbcompliance.view_reviewcompliancetask"]),
    add=MenuItem(
        label=_("Add Report Indicator"),
        endpoint="wbcompliance:reviewcompliancetask-list",
        permission=ItemPermission(method=is_admin_compliance, permissions=["wbcompliance.add_reviewcompliancetask"]),
    ),
)


COMPLIANCEREPORT_MENU = Menu(
    label=_("Reports"),
    items=[
        REVIEWCOMPLIANCETASK_MENUITEM,
        COMPLIANCEACTION_MENUITEM,
        ADMIN_COMPLIANCEEVENT_MENUITEM,
        COMPLIANCETASKMATRIX_MENUITEM,
    ],
)

COMPLIANCE_SETTINGS_MENU = Menu(
    label=_("Settings"),
    items=[
        COMPLIANCETYPE_MENUITEM,
        COMPLIANCETASKGROUP_MENUITEM,
        COMPLIANCETASK_MENUITEM,
        COMPLIANCETASKINSTANCE_MENUITEM,
    ],
)



from .rules import (
    RuleGroupRepresentationSerializer,
    RiskRuleModelSerializer,
    RiskRuleRepresentationSerializer,
    RuleBackendModelSerializer,
    RuleBackendRepresentationSerializer,
    RuleCheckedObjectRelationshipModelSerializer,
    RuleCheckedObjectRelationshipRepresentationSerializer,
    RuleThresholdRepresentationSerializer,
)
from .checks import RiskCheckModelSerializer, RiskCheckRepresentationSerializer
from .incidents import (
    CheckedObjectIncidentRelationshipModelSerializer,
    CheckedObjectIncidentRelationshipRepresentationSerializer,
    RiskIncidentModelSerializer,
    RiskIncidentRepresentationSerializer,
    RiskIncidentTypeRepresentationSerializer,
    RuleThresholdModelSerializer,
)



from django.db.models import Count, F, OuterRef, Q, Subquery
from django.db.models.functions import Coalesce
from django.utils.translation import gettext_lazy as _
from wbcore import filters as wb_filters
from wbcore.contrib.directory.models import Person

from wbcompliance.models.risk_management import RiskRule, RuleThreshold

from ...models.risk_management.rules import RuleGroup
from .utils import get_default_handler


class RiskRuleFilterSet(wb_filters.FilterSet):
    checked_object_relationships = wb_filters.MultipleChoiceContentTypeFilter(
        label="Checked Objects",
        field_name="checked_object_relationships",
        distinct=True,
        object_id_label="checked_object_relationships__checked_object_id",
        content_type_label="checked_object_relationships__checked_object_content_type",
        hidden=True,
    )

    exclude_informational_rule = wb_filters.BooleanFilter(
        method="filter_exclude_informational_rule", initial=True, label=_("Exclude Informational Rule")
    )
    rule_group = wb_filters.ModelChoiceFilter(
        label=_("Group"),
        queryset=RuleGroup.objects.all(),
        endpoint=RuleGroup.get_representation_endpoint(),
        value_key=RuleGroup.get_representation_value_key(),
        label_key=RuleGroup.get_representation_label_key(),
        method="filter_rule_group",
    )

    def filter_rule_group(self, queryset, label, value):
        if value:
            return queryset.filter(Q(rule_backend__rule_group=value) | Q(rule_backend__rule_group__isnull=True))
        return queryset

    def filter_exclude_informational_rule(self, queryset, label, value):
        if value:
            queryset = queryset.annotate(
                threshold_count=Coalesce(
                    Subquery(
                        RuleThreshold.objects.filter(rule=OuterRef("pk"))
                        .values("rule")
                        .annotate(c=Count("rule"))
                        .values("c")[:1]
                    ),
                    0,
                ),
                informational_count=Coalesce(
                    Subquery(
                        RuleThreshold.objects.filter(rule=OuterRef("pk"), severity__is_informational=True)
                        .values("rule")
                        .annotate(c=Count("rule"))
                        .values("c")[:1]
                    ),
                    0,
                ),
            )
            return queryset.exclude(Q(threshold_count=F("informational_count")) & ~Q(threshold_count=0))
        return queryset

    in_breach = wb_filters.ChoiceFilter(
        method="filter_in_breach",
        label=_("Open Incidents"),
        choices=[("BREACH", "In Breach"), ("PASSED", "Passed"), ("INACTIVE", "Inactive")],
    )
    open_incidents_count = wb_filters.NumberFilter(
        label=_("Opened Incident"), field_name="open_incidents_count", lookup_expr="exact"
    )
    open_incidents_count__lte = wb_filters.NumberFilter(
        label=_("Opened Incident"), field_name="open_incidents_count", lookup_expr="lte"
    )
    open_incidents_count__gte = wb_filters.NumberFilter(
        label=_("Opened Incident"), field_name="open_incidents_count", lookup_expr="gte"
    )
    notified_by = wb_filters.ModelChoiceFilter(
        label=_("Notifed Users"),
        initial=get_default_handler,
        queryset=Person.objects.all(),
        endpoint=Person.get_representation_endpoint(),
        value_key=Person.get_representation_value_key(),
        label_key=Person.get_representation_label_key(),
        method="filter_notified_by",
    )

    def filter_notified_by(self, queryset, name, value):
        if value:
            thresholds = RuleThreshold.objects.filter(
                Q(notifiable_users=value) | Q(notifiable_groups__in=value.user_account.groups.all())
            )
            return queryset.filter(thresholds__in=thresholds).distinct()
        return queryset

    def filter_in_breach(self, queryset, name, value):
        if value:
            return queryset.filter(in_breach=value)
        return queryset

    class Meta:
        model = RiskRule
        fields = {
            "rule_backend": ["exact"],
            "is_enable": ["exact"],
            "only_passive_check_allowed": ["exact"],
            "is_silent": ["exact"],
            "is_mandatory": ["exact"],
            "automatically_close_incident": ["exact"],
        }



# Generated by Django 5.0.14 on 2025-06-26 15:06
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('wbcompliance', '0020_rename_trigger_content_type_riskcheck_checked_object_content_type_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='riskcheck',
            name='activator_content_type',
            field=models.ForeignKey(blank=True, max_length=256, null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype'),
        ),
        migrations.AddField(
            model_name='riskcheck',
            name='activator_id',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
    ]



from django.urls import include, path
from wbcore.routers import WBCoreRouter

from wbcompliance import viewsets

router = WBCoreRouter()
router.register(
    r"complianceformrepresentation",
    viewsets.ComplianceFormRepresentationViewSet,
    basename="complianceformrepresentation",
)
router.register(
    r"complianceformsectionrepresentation",
    viewsets.ComplianceFormSectionRepresentationViewSet,
    basename="complianceformsectionrepresentation",
)
router.register(
    r"complianceformsignaturesectionrepresentation",
    viewsets.ComplianceFormSignatureSectionRepresentationViewSet,
    basename="complianceformsignaturesectionrepresentation",
)

router.register(
    r"complianceformtyperepresentation",
    viewsets.ComplianceFormTypeRepresentationViewSet,
    basename="complianceformtyperepresentation",
)
router.register(r"complianceformtype", viewsets.ComplianceFormTypeViewSet, basename="complianceformtype")

router.register(r"complianceform", viewsets.ComplianceFormModelViewSet)
router.register(
    r"complianceformsignature", viewsets.ComplianceFormSignatureModelViewSet, basename="complianceformsignature"
)

router.register(r"complianceformsection", viewsets.ComplianceFormSectionViewSet, basename="complianceformsection")

router.register(r"complianceformrule", viewsets.ComplianceFormRuleViewSet, basename="complianceformrule")

router.register(
    r"compliancetaskrepresentation",
    viewsets.ComplianceTaskRepresentationViewSet,
    basename="compliancetaskrepresentation",
)
router.register(r"compliancetask", viewsets.ComplianceTaskModelViewSet, basename="compliancetask")
router.register(
    r"compliancetaskinstance", viewsets.ComplianceTaskInstanceModelViewSet, basename="compliancetaskinstance"
)

router.register(r"complianceaction", viewsets.ComplianceActionModelViewSet, basename="complianceaction")
router.register(r"complianceevent", viewsets.ComplianceEventModelViewSet, basename="complianceevent")

router.register(
    r"compliancetyperepresentation",
    viewsets.ComplianceTypeRepresentationViewSet,
    basename="compliancetyperepresentation",
)
router.register(r"compliancetype", viewsets.ComplianceTypeModelViewSet, basename="compliancetype")

router.register(
    r"reviewcompliancetaskrepresentation",
    viewsets.ReviewComplianceTaskRepresentationViewSet,
    basename="reviewcompliancetaskrepresentation",
)
router.register(r"reviewcompliancetask", viewsets.ReviewComplianceTaskModelViewSet, basename="reviewcompliancetask")


router.register(
    r"compliancetaskgrouprepresentation",
    viewsets.ComplianceTaskGroupRepresentationViewSet,
    basename="compliancetaskgrouprepresentation",
)
router.register(r"compliancetaskgroup", viewsets.ComplianceTaskGroupModelViewSet, basename="compliancetaskgroup")

# Subrouter for compliance type
compliance_type_router = WBCoreRouter()
compliance_type_router.register(
    r"complianceaction",
    viewsets.TypeComplianceActionModelViewSet,
    basename="type-complianceaction",
)
compliance_type_router.register(
    r"complianceevent",
    viewsets.TypeComplianceEventModelViewSet,
    basename="type-complianceevent",
)


# Subrouter for forms to be signed of a Compliance Form
compliance_form_router = WBCoreRouter()
compliance_form_router.register(
    r"signatures",
    viewsets.CFComplianceFormSignatureModelViewSet,
    basename="complianceform-signatures",
)
# Subrouter for the section of a Compliance Form
compliance_form_router.register(
    r"sections",
    viewsets.ComplianceFormSectionComplianceFormViewSet,
    basename="complianceform-sections",
)


# Subrouter for the rules of a section
section_router = WBCoreRouter()
section_router.register(
    r"rules",
    viewsets.ComplianceFormSectionRuleViewSet,
    basename="complianceformsection-rules",
)

section_router.register(
    r"signauturerules",
    viewsets.ComplianceFormSignatureSectionRuleViewSet,
    basename="complianceformsignaturesection-rules",
)


# Subrouter for the compliance task of a group
taskgroup_router = WBCoreRouter()
taskgroup_router.register(
    r"compliancetaskgroup",
    viewsets.ComplianceTaskComplianceTaskGroupModelViewSet,
    basename="compliancetaskgroup-compliancetask",
)

task_router = WBCoreRouter()
task_router.register(
    r"compliancetask",
    viewsets.ComplianceTaskInstanceComplianceTaskModelViewSet,
    basename="compliancetask-compliancetaskinstance",
)


# Subrouter for the compliance task of a review
review_router = WBCoreRouter()
review_router.register(
    r"compliancetasknogroup",
    viewsets.ComplianceTaskReviewNoGroupModelViewSet,
    basename="review-compliancetasknogroup",
)
review_router.register(
    r"compliancetaskinstancenogroup",
    viewsets.ComplianceTaskInstanceReviewNoGroupModelViewSet,
    basename="review-compliancetaskinstancenogroup",
)

# Subrouter for the compliance task of a group of  a review
review_group_router = WBCoreRouter()
review_group_router.register(
    r"compliancetaskgroup",
    viewsets.ComplianceTaskReviewGroupModelViewSet,
    basename="review-compliancetaskgroup",
)
review_group_router.register(
    r"compliancetaskinstancegroup",
    viewsets.ComplianceTaskInstanceReviewGroupModelViewSet,
    basename="review-compliancetaskinstancegroup",
)

# Risk Management url routing

router.register("riskcheckrepresentation", viewsets.RiskCheckRepresentationViewSet, basename="riskcheckrepresentation")
router.register("rulegrouprepresentation", viewsets.RuleGroupRepresentationViewSet, basename="rulegrouprepresentation")
router.register(
    "riskincidentrepresentation", viewsets.RiskIncidentRepresentationViewSet, basename="riskincidentrepresentation"
)
router.register(
    "riskincidenttyperepresentation",
    viewsets.RiskIncidentTypeRepresentationViewSet,
    basename="riskincidenttyperepresentation",
)
router.register(
    "checkedobjectincidentrelationshiprepresentation",
    viewsets.CheckedObjectIncidentRelationshipRepresentationViewSet,
    basename="checkedobjectincidentrelationshiprepresentation",
)
router.register(
    "rulechecked_objectrelationshiprepresentation",
    viewsets.RuleCheckedObjectRelationshipRepresentationViewSet,
    basename="rulechecked_objectrelationshiprepresentation",
)
router.register(
    "rulebackendrepresentation", viewsets.RuleBackendRepresentationViewSet, basename="rulebackendrepresentation"
)
router.register(
    "rulethresholdrepresentation", viewsets.RuleThresholdRepresentationViewSet, basename="rulethresholdrepresentation"
)
router.register("riskrulerepresentation", viewsets.RiskRuleRepresentationViewSet, basename="riskrulerepresentation")
router.register("riskincident", viewsets.RiskIncidentModelViewSet, basename="riskincident")
router.register(
    "checkedobjectincidentrelationship",
    viewsets.CheckedObjectIncidentRelationshipModelViewSet,
    basename="checkedobjectincidentrelationship",
)
router.register("riskcheck", viewsets.RiskCheckModelViewSet, basename="riskcheck")
router.register("riskrule", viewsets.RiskRuleModelViewSet, basename="riskrule")

rule_group_router = WBCoreRouter()
rule_group_router.register(
    r"incident",
    viewsets.RiskIncidentRiskRuleModelViewSet,
    basename="riskrule-incident",
)
rule_group_router.register(
    r"relationship",
    viewsets.RuleCheckedObjectRelationshipRiskRuleModelViewSet,
    basename="riskrule-relationship",
)
rule_group_router.register(
    r"threshold",
    viewsets.RuleThresholdRiskRuleModelViewSet,
    basename="riskrule-threshold",
)

incident_group_router = WBCoreRouter()
incident_group_router.register(
    r"relationship",
    viewsets.CheckedObjectIncidentRelationshipRiskRuleModelViewSet,
    basename="riskincident-relationship",
)

# Pandas ViewSet
router.register("compliancetaskmatrix", viewsets.ComplianceTaskMatrixPandasViewSet, basename="compliancetaskmatrix")
router.register(
    "riskmanagementincidenttable", viewsets.RiskManagementIncidentTableView, basename="riskmanagementincidenttable"
)

urlpatterns = [
    path("", include(router.urls)),
    path("riskrule/<int:rule_id>/", include(rule_group_router.urls)),
    path("riskincident/<int:incident_id>/", include(incident_group_router.urls)),
    path("complianceform/<int:compliance_form_id>/", include(compliance_form_router.urls)),
    path("section/<int:section_id>/", include(section_router.urls)),
    path("taskgroup/<int:group_id>/", include(taskgroup_router.urls)),
    path("task/<int:task_id>/", include(task_router.urls)),
    path("review/<int:review_id>/", include(review_router.urls)),
    path("review/<int:review_id>/group/<int:group_id>/", include(review_group_router.urls)),
    path("type/<int:type_id>/", include(compliance_type_router.urls)),
]



from typing import Optional

from django.shortcuts import get_object_or_404
from django.utils.translation import gettext as _
from wbcore.contrib.color.enums import WBColor
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

from wbcompliance.models import (
    ComplianceForm,
    ComplianceFormSection,
    ComplianceFormSignatureSection,
    ComplianceFormType,
)


# TYPE OF THE COMPLIANCE FORM
class ComplianceFormTypeDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        fields = [
            dp.Field(key="name", label=_("Name")),
            dp.Field(key="type", label=_("Type")),
        ]
        return dp.ListDisplay(
            fields=fields,
        )

    def get_instance_display(self) -> Display:
        return create_simple_display([["name", "type"]])


# SECTION OF THE COMPLIANCE FORM
class ComplianceFormSectionDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        fields = [
            dp.Field(key="name", label=_("Titles")),
        ]
        return dp.ListDisplay(
            fields=fields,
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["compliance_form", "name"] if "compliance_form_id" not in self.view.kwargs else ["name", "name"],
                [repeat_field(2, "rules_section")],
            ],
            [create_simple_section("rules_section", _("Rules"), [["rules"]], "rules", collapsed=False)],
        )


# RULES OF THE SECTION
class ComplianceFormRuleDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        fields = [
            dp.Field(key="text", label=_("Text")),
            dp.Field(key="ticked", label=_("Expected Answer")),
        ]
        return dp.ListDisplay(
            fields=fields,
        )

    def get_instance_display(self) -> Display:
        return create_simple_display([["ticked", repeat_field(2, "text")]])


# RULES OF THE SECTION OF THE COMPLIANCE FORM SIGNATURE
class ComplianceFormSignatureSectionRuleDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        fields = [
            dp.Field(key="text", label=_("Text")),
            dp.Field(
                key="ticked", label=_("Answer"), formatting_rules=[dp.FormattingRule(style={"fontWeight": "bold"})]
            ),
            dp.Field(key="comments", label=_("Comments")),
        ]
        if self.request.user.has_perm("wbcompliance.administrate_compliance") and (
            section_id := self.view.kwargs.get("section_id")
        ):
            section_signature = get_object_or_404(ComplianceFormSignatureSection, pk=section_id)
            if section_signature.compliance_form_signature.person != self.request.user.profile:
                fields = [
                    dp.Field(key="text", label=_("Text")),
                    dp.Field(key="expected_result", label=_("Expected Answer")),
                    dp.Field(
                        key="ticked",
                        label=_("Answer"),
                        formatting_rules=[dp.FormattingRule(style={"fontWeight": "bold"})],
                    ),
                    dp.Field(key="comments", label=_("Comments")),
                ]
        return dp.ListDisplay(
            fields=fields,
            formatting=[
                dp.Formatting(
                    column="same_answer",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.YELLOW_LIGHT.value},
                            condition=("==", False),
                        ),
                    ],
                ),
            ],
        )

    def get_instance_display(self) -> Display:
        grid_fields = [
            ["section", "ticked"],
            [repeat_field(2, "text")],
            [repeat_field(2, "comments")],
        ]
        if self.request.user.has_perm("wbcompliance.administrate_compliance") and (
            section_id := self.view.kwargs.get("section_id")
        ):
            section_signature = get_object_or_404(ComplianceFormSignatureSection, pk=section_id)
            if section_signature.compliance_form_signature.person != self.request.user.profile:
                grid_fields = [
                    ["section", "ticked", "expected_result"],
                    [repeat_field(3, "text")],
                    [repeat_field(3, "comments")],
                ]
        return create_simple_display(grid_fields)


# COMPLIANCE FORM
class ComplianceFormDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        fields = [
            dp.Field(key="title", label=_("Title")),
            dp.Field(key="creator", label=_("Creator")),
            dp.Field(key="changer", label=_("Changer")),
            dp.Field(key="changed", label=_("Changed")),
            dp.Field(key="start", label=_("Start")),
            dp.Field(key="end", label=_("End")),
            dp.Field(key="version", label=_("Last Version")),
        ]
        if self.request.user.has_perm("wbcompliance.administrate_compliance"):
            fields.append(dp.Field(key="current_signed", label=_("Current Signed")))
        else:
            fields.append(
                dp.Field(
                    key="is_signed",
                    label=_("Is Signed"),
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"color": WBColor.RED.value, "fontWeight": "bold"},
                        ),
                    ],
                )
            )

        fields.append(dp.Field(key="form_type", label=_("Form Type")))
        fields.append(dp.Field(key="compliance_type", label=_("Administrator")))

        return dp.ListDisplay(
            fields=fields,
            formatting=[
                dp.Formatting(
                    column="status",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.BLUE_LIGHT.value},
                            condition=("==", ComplianceForm.Status.DRAFT.name),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.YELLOW_LIGHT.value},
                            condition=("==", ComplianceForm.Status.ACTIVATION_REQUESTED.name),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                            condition=("==", ComplianceForm.Status.ACTIVE.name),
                        ),
                    ],
                ),
            ],
            legends=[
                dp.Legend(
                    key="status",
                    items=[
                        dp.LegendItem(
                            icon=WBColor.BLUE_LIGHT.value,
                            label=ComplianceForm.Status.DRAFT.label,
                            value=ComplianceForm.Status.DRAFT.value,
                        ),
                        dp.LegendItem(
                            icon=WBColor.YELLOW_LIGHT.value,
                            label=ComplianceForm.Status.ACTIVATION_REQUESTED.label,
                            value=ComplianceForm.Status.ACTIVATION_REQUESTED.value,
                        ),
                        dp.LegendItem(
                            icon=WBColor.GREEN_LIGHT.value,
                            label=ComplianceForm.Status.ACTIVE.label,
                            value=ComplianceForm.Status.ACTIVE.value,
                        ),
                    ],
                ),
            ],
        )

    def get_instance_display(self) -> Display:
        fields = [
            ["form_type", "compliance_type"],
            [repeat_field(2, "title")],
            ["start", "end"],
            ["version", "status"],
            [repeat_field(2, "signers_section")],
        ]
        sections = [
            create_simple_section(
                "signers_section",
                _("Signers"),
                [["assigned_to", "assigned_to", "only_internal"]],
                collapsed=True,
            )
        ]
        if (obj_id := self.view.kwargs.get("pk", None)) and (instance := self.view.get_object()):
            if instance.form_type.type == ComplianceFormType.Type.TEXT:
                fields += [[repeat_field(2, "policy")]]

            elif instance.form_type.type == ComplianceFormType.Type.FORM:
                fields.append([repeat_field(2, "sections_section")])
                sections.append(
                    create_simple_section(
                        "sections_section", _("Sections"), [["sections"]], "sections", collapsed=True
                    )
                )

                total = ComplianceFormSection.objects.filter(compliance_form=obj_id).count()
                for count, section in enumerate(
                    ComplianceFormSection.objects.filter(compliance_form=obj_id).order_by("id")
                ):
                    key = f"rules{section.id}"
                    fields.append([repeat_field(2, f"{key}_section")])
                    sections.append(
                        create_simple_section(
                            f"{key}_section", f"({count + 1}/{total}). {section.name}", [[key]], key, collapsed=True
                        )
                    )

        return create_simple_display(fields, sections)


# COMPLIANCE FORM SIGNATURE
class ComplianceFormSignatureDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        fields = [
            dp.Field(key="compliance_form", label=_("Compliance Form")),
            dp.Field(key="version", label=_("Version")),
            dp.Field(key="start", label=_("Start")),
            dp.Field(key="end", label=_("End")),
            dp.Field(key="signed", label=_("Signed")),
            # dp.Field(key="is_signed",  label=_("Is Signed"))
        ]
        if self.request.user.has_perm("wbcompliance.administrate_compliance"):
            fields.append(dp.Field(key="person", label=_("Signer")))

        return dp.ListDisplay(
            fields=fields,
            formatting=[
                dp.Formatting(
                    column="is_signed",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.YELLOW_LIGHT.value},
                            condition=("==", False),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                            condition=("==", True),
                        ),
                    ],
                ),
            ],
            legends=[
                dp.Legend(
                    # key="is_signed",
                    items=[
                        dp.LegendItem(
                            icon=WBColor.YELLOW_LIGHT.value,
                            label=_("Unsigned"),
                            # value=False,
                        ),
                        dp.LegendItem(
                            icon=WBColor.GREEN_LIGHT.value,
                            label=_("Signed"),
                            # value=True,
                        ),
                    ],
                ),
            ],
        )

    def get_instance_display(self) -> Display:
        fields = [["compliance_form", "version"], ["person", "signed"], ["start", "end"]]
        sections = []
        if obj_id := self.view.kwargs.get("pk", None):
            instance = self.view.get_object()
            if instance.compliance_form.form_type.type == ComplianceFormType.Type.TEXT:
                fields.append([repeat_field(2, "policy")])
            elif instance.compliance_form.form_type.type == ComplianceFormType.Type.FORM:
                total = ComplianceFormSignatureSection.objects.filter(compliance_form_signature=obj_id).count()
                for count, section in enumerate(
                    ComplianceFormSignatureSection.objects.filter(compliance_form_signature=obj_id).order_by("id")
                ):
                    key = f"rules{section.id}"
                    fields.append([repeat_field(2, f"{key}_section")])
                    sections.append(
                        create_simple_section(
                            f"{key}_section", f"({count + 1}/{total}). {section.name}", [[key]], key, collapsed=True
                        )
                    )

        fields.append([repeat_field(2, "remark")])
        return create_simple_display(fields, sections)



from django.apps import apps
from django.db.models.signals import pre_migrate
from pytest_factoryboy import register
from wbcompliance.factories import (
    ComplianceActionFactory,
    ComplianceEventFactory,
    ComplianceFormFactory,
    ComplianceFormRuleFactory,
    ComplianceFormSectionFactory,
    ComplianceFormSignatureFactory,
    ComplianceFormTypeFactory,
    ComplianceTaskFactory,
    ComplianceTaskGroupFactory,
    ComplianceTaskInstanceFactory,
    ComplianceTypeFactory,
    UnsignedComplianceFormSignatureFactory,
)
from wbcompliance.factories.risk_management import (
    CheckedObjectIncidentRelationshipFactory,
    RiskCheckFactory,
    RiskIncidentFactory,
    RiskIncidentTypeFactory,
    RiskRuleFactory,
    RuleBackendFactory,
    RuleCheckedObjectRelationshipFactory,
    RuleThresholdFactory,
)
from wbcore.contrib.authentication.factories import (
    AuthenticatedPersonFactory,
    UserFactory,
)
from wbcore.contrib.directory.factories import PersonFactory

from wbcore.tests.conftest import *  # isort:skip

register(ComplianceFormRuleFactory)
register(ComplianceFormSectionFactory)
register(ComplianceFormTypeFactory)
register(ComplianceFormFactory)
register(ComplianceFormSignatureFactory)
register(UnsignedComplianceFormSignatureFactory)
register(ComplianceTaskFactory)
register(ComplianceTaskInstanceFactory)
register(ComplianceActionFactory)
register(ComplianceEventFactory)
register(ComplianceTypeFactory)
register(ComplianceTaskGroupFactory)
register(RiskCheckFactory)
register(RiskIncidentFactory)
register(CheckedObjectIncidentRelationshipFactory)
register(RuleBackendFactory)
register(RiskRuleFactory)
register(RuleThresholdFactory)
register(RuleCheckedObjectRelationshipFactory)
register(RiskIncidentTypeFactory)
register(PersonFactory)
register(UserFactory)
register(AuthenticatedPersonFactory)


from .signals import *

pre_migrate.connect(app_pre_migration, sender=apps.get_app_config("wbcompliance"))



from typing import Optional

from django.utils.translation import gettext as _
from rest_framework.reverse import reverse
from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

from wbcompliance.models.risk_management import RiskIncident, RiskIncidentType


def get_formatting(id_field="id", color=False):
    formattings = []
    for _type in RiskIncidentType.objects.all():
        style = {"backgroundColor": _type.color}
        if color:
            style["color"] = _type.color
        formattings.append(
            dp.FormattingRule(
                style=style,
                condition=("==", getattr(_type, id_field)),
            ),
        )
    return dp.Formatting(column="severity", formatting_rules=formattings)


def get_legends():
    legends = []
    for _type in RiskIncidentType.objects.all():
        legends.append(dp.LegendItem(icon=_type.color, label=_type.name, value=_type.id))
    return dp.Legend(items=legends, key="severity")


class RiskIncidentDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        fields = [
            dp.Field(key="date_range", label=_("Date Range"), pinned="left"),
            dp.Field(
                key=None,
                label="Rule",
                open_by_default=False,
                children=[
                    # dp.Field(key="status_icon", label=""),
                    dp.Field(key="status", label=_("Status"), show="open"),
                    dp.Field(key="object_repr", label=_("Object")),
                    dp.Field(key="rule", label=_("Rule")),
                    dp.Field(key="threshold_repr", label=_("Breached Threshold")),
                ],
            ),
            dp.Field(
                key=None,
                label="Incident",
                open_by_default=False,
                children=[
                    dp.Field(key="breached_value", label=_("Breached Value")),
                    dp.Field(key="report", label=_("Report Detail"), show="open"),
                    dp.Field(key="checked_date", label=_("Date"), show="open"),
                ],
            ),
            dp.Field(
                key=None,
                label="Management",
                open_by_default=False,
                children=[
                    dp.Field(key="resolved_by", label=_("Resolved By")),
                    dp.Field(key="is_notified", label=_("Users notified"), show="open"),
                    dp.Field(key="ignore_until", label=_("Ignored Until"), show="open"),
                    dp.Field(key="comment", label=_("Comment"), show="open"),
                ],
            ),
        ]
        return dp.ListDisplay(
            fields=fields,
            tree=True,
            tree_group_level_options=[
                dp.TreeGroupLevelOption(
                    filter_key="incident",
                    filter_depth=1,
                    clear_filter=False,
                    filter_whitelist=["checked_object_relationships"],
                    list_endpoint=reverse(
                        "wbcompliance:checkedobjectincidentrelationship-list",
                        args=[],
                        request=self.request,
                    ),
                )
            ],
            # tree_group_lookup="id_repr",
            tree_group_field="date_range",
            legends=[
                get_legends(),
                dp.Legend(
                    key="status",
                    items=[
                        dp.LegendItem(
                            icon=WBIcon.VIEW.icon,
                            label=_("{} Incidents").format(RiskIncident.Status.OPEN.label),
                            value=RiskIncident.Status.OPEN.name,
                        ),
                    ],
                ),
            ],
            formatting=[
                get_formatting(),
            ],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                [repeat_field(3, "status")],
                ["date_range", "rule", "severity"],
                ["breached_content_type", "breached_object_id", "breached_object_repr"],
                ["comment", "resolved_by", "ignore_until"],
                [repeat_field(3, "relationships_section")],
            ],
            [
                create_simple_section(
                    "relationships_section", _("Relationships"), [["relationships"]], "relationships", collapsed=False
                )
            ],
        )


class CheckedObjectIncidentRelationshipRiskRuleDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="status", label=_("status")),
                dp.Field(key="checked_date", label=_("Date")),
                dp.Field(key="rule_check", label=_("Check")),
                dp.Field(key="checked_object", label=_("Checked Object")),
                dp.Field(key="breached_value", label=_("Breached Value")),
                dp.Field(key="report", label=_("Report")),
                dp.Field(key="resolved_by", label=_("resolved_by")),
                dp.Field(key="comment", label=_("comment")),
            ],
            legends=[get_legends()],
            formatting=[
                get_formatting(),
            ],
        )



# Generated by Django 5.0.9 on 2024-10-11 11:47

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbcompliance", "0016_checkedobjectincidentrelationship_report_details_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="rulebackend",
            name="incident_report_template",
            field=models.TextField(
                default='<ul style="line-height: 13px; text-align: left; margin: 0; padding: 10px;">\n    {% for key, value in report_details.items %}\n        <li><strong>{{ key }}: </strong> {{ value | safe }}</li>\n    {% endfor %}\n</ul>\n',
                verbose_name="Incident Template",
            ),
        ),
    ]



from collections import defaultdict

from django.db.models.signals import *


class DisableSignals(object):
    def __init__(self, disabled_signals=None):
        self.stashed_signals = defaultdict(list)
        self.disabled_signals = disabled_signals or [
            pre_init,
            post_init,
            pre_save,
            post_save,
            pre_delete,
            post_delete,
            pre_migrate,
            post_migrate,
        ]

    def __enter__(self):
        for signal in self.disabled_signals:
            self.disconnect(signal)

    def __exit__(self, exc_type, exc_val, exc_tb):
        for signal in list(self.stashed_signals):
            self.reconnect(signal)

    def disconnect(self, signal):
        self.stashed_signals[signal] = signal.receivers
        signal.receivers = []

    def reconnect(self, signal):
        signal.receivers = self.stashed_signals.get(signal, [])
        del self.stashed_signals[signal]


# https://stackoverflow.com/questions/55578230/django-how-to-visualize-signals-and-save-overrides
# RECEIVER_MODELS = re.compile(r"sender=(\w+)\W")


# class DisableSignalsNotification_OLD_VERSION(DisableSignals):
#     def __enter__(self):
#         for signal in self.disabled_signals:
#             if not isinstance(signal, ModelSignal):
#                 continue
#             for _, receiver in signal.receivers:
#                 rcode = inspect.getsource(receiver())
#                 rmodel = RECEIVER_MODELS.findall(rcode)
#                 if "Notification" in rmodel:
#                     self.disconnect(signal)


class temp_disconnect_signal:
    """Temporarily disconnect a model from a signal"""

    def __init__(self, signal, receiver, sender, dispatch_uid=None):
        self.signal = signal
        self.receiver = receiver
        self.sender = sender
        self.dispatch_uid = dispatch_uid

    def __enter__(self):
        self.signal.disconnect(
            receiver=self.receiver,
            sender=self.sender,
            dispatch_uid=self.dispatch_uid,
        )

    def __exit__(self, type, value, traceback):
        self.signal.connect(
            receiver=self.receiver,
            sender=self.sender,
            dispatch_uid=self.dispatch_uid,
        )


class DisableSignalsNotification(temp_disconnect_signal):
    def __init__(self, dispatch_uid=None):
        self.signal = post_save
        self.receiver = post_create_notification
        self.sender = Notification
        self.dispatch_uid = dispatch_uid



from wbcore import filters as wb_filters

from wbcompliance.models.risk_management import RiskCheck


class RiskCheckFilterSet(wb_filters.FilterSet):
    checked_objects = wb_filters.MultipleChoiceContentTypeFilter(
        label="Triggerers",
        field_name="checked_objects",
        object_id_label="checked_object_id",
        content_type_label="checked_object_content_type",
        distinct=True,
        hidden=True,
    )
    passive_check = wb_filters.BooleanFilter(initial=True, required=True, label="Passive")

    class Meta:
        model = RiskCheck

        fields = {
            "rule": ["exact"],
            "creation_datetime": ["gte", "exact", "lte"],
            "evaluation_date": ["exact"],
            "status": ["exact"],
        }



import factory
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType

from wbcompliance.models.risk_management.checks import RiskCheck


class RiskCheckFactory(factory.django.DjangoModelFactory):
    rule = factory.SubFactory("wbcompliance.factories.risk_management.rules.RiskRuleFactory")
    evaluation_date = factory.Faker("date_object")
    checked_object_content_type = factory.LazyAttribute(lambda o: ContentType.objects.get_for_model(get_user_model()))
    checked_object_id = 1

    @factory.post_generation
    def rule_checked_object_relationship(self, create, extracted, **kwargs):
        if not create:
            return
        if extracted:
            self.rule = extracted.rule
            self.checked_object_id = extracted.checked_object_id
            self.checked_object_content_type = extracted.checked_object_content_type

    class Meta:
        model = RiskCheck



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class RuleThresholdRiskRuleEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        if rule_id := self.view.kwargs.get("rule_id"):
            return reverse(
                "wbcompliance:riskrule-threshold-list",
                args=[rule_id],
                request=self.request,
            )
        return super().get_endpoint(**kwargs)


class RuleCheckedObjectRelationshipRiskRuleEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        if rule_id := self.view.kwargs.get("rule_id"):
            return reverse(
                "wbcompliance:riskrule-relationship-list",
                args=[rule_id],
                request=self.request,
            )
        return super().get_endpoint(**kwargs)


class RiskRuleEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return super().get_endpoint(**kwargs)

    def get_delete_endpoint(self, **kwargs):
        return None



from datetime import date

import pytest
from dateutil.relativedelta import relativedelta
from rest_framework import status
from rest_framework.test import APIRequestFactory
from wbcore.test.utils import (
    get_data_from_factory,
    get_kwargs,
    get_model_factory,
    get_or_create_superuser,
)

from wbcompliance.viewsets import ComplianceFormModelViewSet

from .mixins import UserTestMixin


@pytest.mark.django_db
class TestSpecificSerializers(UserTestMixin):
    @pytest.mark.parametrize("mvs", [ComplianceFormModelViewSet])
    def test_update_compliance_form_model_serializer(self, mvs, user):
        factory = get_model_factory(mvs().get_serializer_class().Meta.model)
        obj = factory(assigned_to=(user.groups.first(),))
        user = get_or_create_superuser()

        data = get_data_from_factory(obj, mvs, update=True, superuser=user)
        request = APIRequestFactory().put("", data)
        request.user = user

        kwargs = get_kwargs(obj, mvs, request)
        vs = mvs.as_view({"put": "update"})
        response = vs(request, **kwargs, pk=obj.pk)
        assert response.status_code == status.HTTP_200_OK
        assert response.data.get("instance")

    @pytest.mark.parametrize("mvs", [ComplianceFormModelViewSet])
    def test_validate_date_compliance_form(self, mvs):
        user = get_or_create_superuser()
        factory = get_model_factory(mvs().get_serializer_class().Meta.model)
        now = date.today()
        obj = factory(start=now, end=now - relativedelta(months=+5))
        data = get_data_from_factory(obj, mvs, update=True, superuser=user)
        request = APIRequestFactory().post("", data)
        request.user = user
        kwargs = get_kwargs(obj, mvs, request)
        vs = mvs.as_view({"post": "create"})
        response = vs(request, **kwargs)
        assert response.status_code == status.HTTP_400_BAD_REQUEST



from .checks import RiskCheckModelViewSet, RiskCheckRepresentationViewSet
from .incidents import (
    CheckedObjectIncidentRelationshipModelViewSet,
    CheckedObjectIncidentRelationshipRepresentationViewSet,
    CheckedObjectIncidentRelationshipRiskRuleModelViewSet,
    RiskIncidentModelViewSet,
    RiskIncidentRepresentationViewSet,
    RiskIncidentRiskRuleModelViewSet,
    RiskIncidentTypeRepresentationViewSet,
)
from .rules import (
    RuleGroupRepresentationViewSet,
    RiskRuleModelViewSet,
    RiskRuleRepresentationViewSet,
    RuleBackendRepresentationViewSet,
    RuleCheckedObjectRelationshipRepresentationViewSet,
    RuleCheckedObjectRelationshipRiskRuleModelViewSet,
    RuleThresholdRepresentationViewSet,
    RuleThresholdRiskRuleModelViewSet,
)
from .tables import RiskManagementIncidentTableView



import re
from datetime import date, datetime, timedelta

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.db.models import Count, Max, Q, QuerySet
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.template.loader import get_template
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from django_fsm import FSMField, transition
from pandas.tseries.offsets import MonthEnd, YearEnd
from wbcore.contrib.color.enums import WBColor
from wbcore.contrib.icons import WBIcon
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.enums import RequestType
from wbcore.markdown.utils import custom_url_fetcher
from wbcore.metadata.configs.buttons import ActionButton, ButtonDefaultColor
from wbcore.models import WBModel
from wbcore.models.fields import YearField
from wbcore.utils.models import ComplexToStringMixin
from weasyprint import HTML

from .compliance_type import ComplianceDocumentMixin, ComplianceType, can_active_request
from .enums import IncidentSeverity

User = get_user_model()


def can_draft_request(instance, user: "User") -> bool:
    if instance.is_instance:
        return False
    return user.has_perm("wbcompliance.administrate_compliance")


class ComplianceTaskGroup(WBModel):
    name = models.CharField(max_length=255, verbose_name=_("Name"))
    order = models.PositiveIntegerField(null=True, blank=True, verbose_name=_("Order"))

    class Meta:
        verbose_name = "Compliance Task Group"
        verbose_name_plural = "Compliance Task Groups"

    def __str__(self) -> str:
        return "{}".format(self.name)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcompliance:compliancetaskgroup"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcompliance:compliancetaskgrouprepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{name}}"


class ComplianceTask(WBModel):
    class Occurrence(models.TextChoices):
        YEARLY = "YEARLY", "Yearly"
        QUARTERLY = "QUARTERLY", "Quarterly"
        MONTHLY = "MONTHLY", "Monthly"
        NEVER = "NEVER", "Never"

        @classmethod
        def get_color_map(cls) -> list:
            colors = [WBColor.GREEN_LIGHT.value, WBColor.YELLOW_LIGHT.value, WBColor.RED_LIGHT.value]
            return [choice for choice in zip(cls, colors)]

    title = models.CharField(max_length=255, verbose_name=_("Title"))
    description = models.TextField(default="", blank=True, verbose_name=_("Description"))
    occurrence = models.CharField(
        max_length=32,
        default=Occurrence.MONTHLY,
        choices=Occurrence.choices,
        verbose_name=_("Occurrence"),
    )
    active = models.BooleanField(default=True)
    group = models.ForeignKey(
        to="wbcompliance.ComplianceTaskGroup",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="tasks_related",
        verbose_name=_("Group"),
    )
    review = models.ManyToManyField(
        "wbcompliance.ReviewComplianceTask",
        related_name="tasks",
        blank=True,
        verbose_name=_("Review"),
        help_text=_("list of reviews that contain this task"),
    )
    risk_level = models.CharField(
        max_length=32,
        blank=True,
        null=True,
        choices=IncidentSeverity.choices,
        verbose_name=_("Risk Level"),
    )
    remarks = models.TextField(null=True, blank=True, verbose_name=_("Remarks"))
    type = models.ForeignKey(
        to=ComplianceType, on_delete=models.PROTECT, related_name="tasks_of_type", verbose_name=_("Type")
    )

    class Meta:
        verbose_name = "Compliance Task"
        verbose_name_plural = "Compliance Tasks"

    def generate_compliance_task_instance(self, occured: date | None = None, link_instance_review: bool = False):
        kwargs = {"task": self}
        if _instance := ComplianceTaskInstance.objects.filter(task=self).last():
            kwargs.update({"status": _instance.status, "text": _instance.text, "summary_text": _instance.summary_text})
        new_instance = ComplianceTaskInstance.objects.create(**kwargs)

        if occured:
            new_instance.occured = occured
            new_instance.save()

        if link_instance_review:
            for review in self.review.filter(Q(is_instance=False) & Q(review_task=None)):
                qs = ReviewComplianceTask.objects.filter(review_task=review, is_instance=True)
                if review_instance := qs.filter(occured=new_instance.occured).last():
                    new_instance.review.add(review_instance)
        return new_instance

    def __str__(self):
        return "{}".format(self.title)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcompliance:compliancetask"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcompliance:compliancetaskrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{title}}"


class ComplianceTaskInstance(models.Model):
    class Status(models.TextChoices):
        NOT_CHECKED = "NOT_CHECKED", "Not Checked"
        WARNING = "WARNING", "Warning"
        FOR_INFO = "FOR_INFO", "For Info"
        NOTHING_TO_REPORT = "NOTHING_TO_REPORT", "Nothing to Report"
        BREACH = "BREACH", "Breach"

        @classmethod
        def get_color_map(cls) -> list:
            colors = [
                WBColor.GREY.value,
                WBColor.YELLOW_LIGHT.value,
                WBColor.YELLOW.value,
                WBColor.BLUE_LIGHT.value,
                WBColor.RED_LIGHT.value,
            ]
            return [choice for choice in zip(cls, colors)]

    task = models.ForeignKey(
        on_delete=models.CASCADE,
        to="wbcompliance.ComplianceTask",
        related_name="task_instances_related",
        verbose_name=_("Compliance Task"),
    )
    occured = models.DateField(auto_now_add=True, verbose_name=_("Occured"))
    status = models.CharField(
        max_length=32,
        default=Status.NOT_CHECKED,
        choices=Status.choices,
        verbose_name=_("Status"),
    )
    text = models.TextField(default="", blank=True, verbose_name=_("Text"))
    summary_text = models.TextField(null=True, blank=True, verbose_name=_("Summary Text"))
    review = models.ManyToManyField(
        "wbcompliance.ReviewComplianceTask",
        related_name="task_instances",
        blank=True,
        verbose_name=_("Review"),
        help_text=_("list of reviews that contain this task instance"),
    )

    class Meta:
        verbose_name = "Compliance Task Instance"
        verbose_name_plural = "Compliance Task Instances"

    @classmethod
    def get_max_depth(cls) -> int | None:
        return cls.objects.all().values("task").annotate(dcount=Count("task")).aggregate(max=Max("dcount")).get("max")

    @classmethod
    def get_dict_max_count_task(cls) -> dict:
        if cls.get_max_depth():
            return cls.objects.all().values("task").annotate(dcount=Count("task")).latest("dcount")
        return {}

    def __str__(self) -> str:
        return "{}".format(self.task.title)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcompliance:compliancetaskinstance"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{task__title}} : {{id}}"


class ComplianceAction(models.Model):
    class Status(models.TextChoices):
        TO_BE_DONE = "TO_BE_DONE", "To be done"
        WORK_IN_PROGRESS = "WORK_IN_PROGRESS", "Work in Progress"
        DONE = "DONE", "Done"

        @classmethod
        def get_color_map(cls) -> list:
            colors = [WBColor.GREY.value, WBColor.YELLOW_LIGHT.value, WBColor.GREEN_LIGHT.value]
            return [choice for choice in zip(cls, colors)]

    title = models.CharField(max_length=255, verbose_name=_("Title"))
    description = models.TextField(
        default="",
        null=True,
        blank=True,
        verbose_name=_("Description"),
    )
    summary_description = models.TextField(
        null=True,
        blank=True,
        verbose_name=_("Summary Description"),
    )
    deadline = models.DateField(null=True, blank=True, verbose_name=_("Deadline"))
    progress = models.FloatField(
        default=0, validators=[MinValueValidator(0.0), MaxValueValidator(1.0)], verbose_name=_("Progress")
    )
    status = models.CharField(
        max_length=32,
        default=Status.TO_BE_DONE,
        choices=Status.choices,
        verbose_name=_("Status"),
    )
    type = models.ForeignKey(
        to=ComplianceType, on_delete=models.PROTECT, related_name="actions_of_type", verbose_name=_("Type")
    )
    active = models.BooleanField(default=True, verbose_name=_("Active"))
    creator = models.ForeignKey(
        to="directory.Person",
        null=True,
        blank=True,
        related_name="compliance_actions",
        verbose_name=_("Creator"),
        on_delete=models.SET_NULL,
    )
    created = models.DateTimeField(auto_now_add=True, verbose_name=_("Created"))
    changer = models.ForeignKey(
        to="directory.Person",
        null=True,
        blank=True,
        verbose_name=_("Changer"),
        related_name="updated_actions",
        on_delete=models.SET_NULL,
    )
    last_modified = models.DateTimeField(auto_now=True, verbose_name=_("Last modified"))

    class Meta:
        verbose_name = "Compliance Action"
        verbose_name_plural = "Compliance Actions"

    def __str__(self) -> str:
        return "{}".format(self.title)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcompliance:complianceaction"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{title}}"


class ComplianceEvent(models.Model):
    class Type(models.TextChoices):
        INCIDENT = "INCIDENT", "Incident"
        INFO = "INFO", "Info"

        @classmethod
        def get_color_map(cls) -> list:
            colors = [WBColor.GREY.value, WBColor.BLUE_LIGHT.value]
            return [choice for choice in zip(cls, colors)]

    type_event = models.CharField(
        max_length=32,
        default=Type.INCIDENT,
        choices=Type.choices,
        verbose_name=_("Type Event"),
    )
    level = models.CharField(
        max_length=32,
        default=IncidentSeverity.LOW,
        choices=IncidentSeverity.choices,
        verbose_name=_("Level"),
    )
    title = models.CharField(max_length=255, verbose_name=_("Title"))
    exec_summary = models.TextField(
        null=True,
        blank=True,
        verbose_name=_("Executive Summary"),
    )
    exec_summary_board = models.TextField(
        null=True,
        blank=True,
        verbose_name=_("Executive Summary for the Board"),
    )
    description = models.TextField(default="", blank=True, verbose_name=_("Description"))
    actions_taken = models.TextField(default="", blank=True, verbose_name=_("Actions Taken"))
    consequences = models.TextField(default="", blank=True, verbose_name=_("Consequences"))
    future_suggestions = models.TextField(default="", blank=True, verbose_name=_("Future Suggestions"))
    type = models.ForeignKey(
        to=ComplianceType, on_delete=models.PROTECT, related_name="events_of_type", verbose_name=_("Type")
    )
    active = models.BooleanField(default=True)
    creator = models.ForeignKey(
        to="directory.Person",
        null=True,
        blank=True,
        related_name="compliance_events",
        verbose_name=_("Creator"),
        on_delete=models.SET_NULL,
    )
    created = models.DateTimeField(auto_now_add=True, verbose_name=_("Created"))
    changer = models.ForeignKey(
        to="directory.Person",
        null=True,
        blank=True,
        related_name="updated_events",
        verbose_name=_("Changer"),
        on_delete=models.SET_NULL,
    )
    last_modified = models.DateTimeField(auto_now=True, verbose_name=_("Last modified"))
    confidential = models.BooleanField(default=False, verbose_name=_("Confidential"))

    class Meta:
        verbose_name = "Compliance Event"
        verbose_name_plural = "Compliance Events"

        notification_types = [
            create_notification_type(
                code="wbcompliance.complianceevent.notify",
                title="Compliance Event Notification",
                help_text="Sends out a notification when a new compliance event was created.",
                email=True,
                web=False,
                mobile=False,
                is_lock=True,
            )
        ]

    def __str__(self) -> str:
        return "{}".format(self.title)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcompliance:complianceevent"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{title}}"


class ReviewComplianceTask(ComplianceDocumentMixin, ComplexToStringMixin, WBModel):
    class Occurrence(models.TextChoices):
        YEARLY = "YEARLY", "Yearly"
        QUARTERLY = "QUARTERLY", "Quarterly"
        MONTHLY = "MONTHLY", "Monthly"
        NEVER = "NEVER", "Never"

    class Status(models.TextChoices):
        DRAFT = "DRAFT", "Draft"
        VALIDATION_REQUESTED = "VALIDATION_REQUESTED", "Validation Requested"
        VALIDATED = "VALIDATED", "Validated"

        @classmethod
        def get_color_map(cls) -> list:
            colors = [
                WBColor.BLUE_LIGHT.value,
                WBColor.YELLOW_LIGHT.value,
                WBColor.GREEN_LIGHT.value,
            ]
            return [choice for choice in zip(cls, colors)]

    class Meta:
        verbose_name = "Review Compliance Task"
        verbose_name_plural = "Review Compliance Tasks"

        notification_types = [
            create_notification_type(
                code="wbcompliance.reviewcompliancetask.notify",
                title="Compliance Task Review Notification",
                help_text="Notifies you when a compliance task can be reviewed",
                email=True,
                web=False,
                mobile=False,
                is_lock=True,
            )
        ]

    title = models.CharField(max_length=255, verbose_name=_("Title"))
    from_date = models.DateField(null=True, blank=True, verbose_name=_("From"))
    to_date = models.DateField(null=True, blank=True, verbose_name=_("To"))
    description = models.TextField(default="", blank=True, verbose_name=_("Description"))
    year = YearField(
        validators=[MinValueValidator(1000), MaxValueValidator(9999)], null=True, blank=True, verbose_name=_("Year")
    )
    creator = models.ForeignKey(
        to="directory.Person",
        null=True,
        blank=True,
        verbose_name=_("Creator"),
        related_name="author_review_compliane_tasks",
        on_delete=models.CASCADE,
    )
    created = models.DateTimeField(auto_now_add=True, verbose_name=_("Created"))
    changer = models.ForeignKey(
        "directory.Person", null=True, blank=True, verbose_name=_("Changer"), on_delete=models.deletion.SET_NULL
    )
    changed = models.DateTimeField(auto_now=True, verbose_name=_("Changed"))
    status = FSMField(default=Status.DRAFT, choices=Status.choices, verbose_name=_("Status"))
    occurrence = models.CharField(
        max_length=32,
        default=Occurrence.MONTHLY,
        choices=Occurrence.choices,
        verbose_name=_("Occurrence"),
    )

    is_instance = models.BooleanField(default=False, verbose_name=_("Is occurrence"))
    review_task = models.ForeignKey(
        to="wbcompliance.ReviewComplianceTask",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="review_tasks",
        verbose_name=_("Parent Review"),
    )
    occured = models.DateField(null=True, blank=True, verbose_name=_("Occured Instance"))
    type = models.ForeignKey(
        to=ComplianceType,
        on_delete=models.PROTECT,
        related_name="reviewtask_of_type",
        verbose_name=_("Type"),
    )

    def notify(self, title, msg, recipients: QuerySet["User"]) -> None:
        for user in recipients:
            send_notification(
                code="wbcompliance.reviewcompliancetask.notify",
                title=title,
                body=msg,
                user=user,
                reverse_name="wbcompliance:reviewcompliancetask-detail",
                reverse_args=[self.id],
            )

    def _remove_styled_paragraph(self, html_content: str) -> str:
        tags = ["p", "span"]
        for tag in tags:
            reg_str = "<" + tag + "(.*?)" + ">"
            reg_style = '(style=".*?;")'
            _attributs = re.findall(reg_str, html_content)
            dict_attrs = {}
            for attribut in _attributs:
                dict_attrs[attribut] = re.sub(reg_style, "", attribut)

            for _key, _value in dict_attrs.items():
                html_content = re.sub(_key, _value, html_content)

        html_content = re.sub("<p>&nbsp;</p>", "", html_content)
        return html_content

    def generate_pdf(self) -> bytes:
        html = get_template("compliance/review_compliance_task_report.html")
        table = {}
        tasks = ComplianceTask.objects.filter(Q(review=self) & Q(active=True)).order_by("group__order")
        group_ids = tasks.values_list("group", flat=True).distinct()
        for group_id in group_ids:
            if group_id:
                group = ComplianceTaskGroup.objects.get(id=group_id)
                table[group.id] = {"name": group.name, "tasks": {}}
            else:
                group_id = ""
                group = None
                table[""] = {"name": "", "tasks": {}}

            for task in tasks.filter(group=group):
                table[group_id]["tasks"][task.id] = {
                    "title": task.title,
                    "description": self._remove_styled_paragraph(task.description),
                    "risk_level": task.risk_level,
                    "remarks": task.remarks,
                }
        html_content = html.render(
            {
                "today": timezone.now(),
                "review": self,
                "table": table,
            }
        )
        return HTML(
            string=html_content, base_url=settings.BASE_ENDPOINT_URL, url_fetcher=custom_url_fetcher
        ).write_pdf()

    def get_period_date(self, today: datetime | None = None):
        if today is None:
            today = timezone.now()
        if self.occurrence == self.Occurrence.YEARLY:
            from_date = date(today.year - 1, 1, 1)
            to_date = from_date + YearEnd(1)
        elif self.occurrence == self.Occurrence.QUARTERLY:
            from_date = (today - MonthEnd(4)).date() + timedelta(days=1)
            to_date = (today - MonthEnd(1)).date()
        elif self.occurrence == self.Occurrence.MONTHLY:
            from_date = (today - MonthEnd(2)).date() + timedelta(days=1)
            to_date = (today - MonthEnd(1)).date()
        else:
            return None, None
        return from_date, to_date

    def generate_review_compliance_task_instance(
        self, current_date: datetime | None = None, link_instance: bool = False, notify_admin: bool = False
    ) -> None:
        """
        allow to generate the occurrence of the Indicators Report.
        current_date: allow to find the year and the period date corresponding to the occurrence. we use today by default
        """
        if current_date is None:
            current_date = timezone.now()
        from_date, to_date = self.get_period_date(current_date)
        date_title = to_date if to_date else current_date
        kwargs = {
            "review_task": self,
            "year": current_date.year,
            "occured": current_date.date(),
            "is_instance": True,
            "status": ReviewComplianceTask.Status.DRAFT,
            "creator": self.creator,
            "changer": self.changer,
            "occurrence": ReviewComplianceTask.Occurrence.NEVER,
            "description": self.description,
            "title": "{} - {}".format(self.title, date_title.strftime("%b %Y")),
            "from_date": from_date,
            "to_date": to_date,
            "type": self.type,
        }
        if _instance := ReviewComplianceTask.objects.filter(Q(review_task=self) & Q(is_instance=True)).last():
            kwargs.update({"creator": _instance.changer, "description": _instance.description})
        new_review = ReviewComplianceTask.objects.create(**kwargs)

        if link_instance:
            for task in ComplianceTask.objects.filter(review__in=[self]):
                if instance := ComplianceTaskInstance.objects.filter(task=task).last():
                    instance.review.add(new_review)

        if notify_admin and (compliance_type := self.type):
            recipients = ComplianceType.get_administrators(compliance_type)
            msg = _("<b>{}</b> is available. You can now complete and validate it.").format(self.title)
            title = _("Instance Indicator Report: {}").format(self.title)
            self.notify(title, msg, recipients)

    def get_task_group_ids_from_review(self, through_task: bool = True, task_with_group: bool = True) -> set:
        if through_task:
            return set(
                self.tasks.filter(group__isnull=not (task_with_group))
                .order_by("group__order")
                .values_list("group", flat=True)
            )
        else:
            return set(
                self.task_instances.filter(task__group__isnull=not (task_with_group))
                .order_by("task__group__order")
                .values_list("task__group", flat=True)
            )

    @transition(
        field=status,
        source=Status.DRAFT,
        target=Status.VALIDATION_REQUESTED,
        permission=lambda _, user: user.has_perm("wbcompliance.administrate_compliance"),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbcompliance:reviewcompliancetask",),
                icon=WBIcon.SEND.icon,
                key="validationrequested",
                label="Request Validation",
                action_label=_("Request Validation"),
                description_fields=_(
                    "<p>Title: <b>{{title}}</b></p>\
                <p>Status: <b>{{status}}</b></p> <p>From: <b>{{from_date}}</b></p>\
                <p>To: <b>{{to_date}}</b></p> <p>Do you want to send this request for validation ?</p>"
                ),
            )
        },
    )
    def validationrequested(self, by=None, description=None, **kwargs):
        if compliance_type := self.type:
            # notify the compliance team without the current user
            if by:
                self.changer = by.profile
            current_user = self.changer if self.changer else self.creator
            recipients = ComplianceType.get_administrators(compliance_type).exclude(profile=current_user)
            msg = _("Validation Request from {} to validate a ompliance Risk Review: <b>{}</b>").format(
                str(current_user), self.title
            )
            title = _("Validation Requested Compliance Risk Review: {}").format(self.title)
            self.notify(title, msg, recipients)

    @transition(
        field=status,
        source=Status.VALIDATION_REQUESTED,
        target=Status.DRAFT,
        permission=lambda _, user: user.has_perm("wbcompliance.administrate_compliance"),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbcompliance:reviewcompliancetask",),
                icon=WBIcon.EDIT.icon,
                key="draft",
                label="Return to Draft Mode",
                action_label=_("Return to Draft Mode"),
                description_fields=_(
                    """
                <p>Title: <b> {{title}} </b></p>
                <p>Status: <b>{{status}}</b></p> <p>From: <b>{{from_date}}</b></p>  <p>To: <b>{{to_date}}</b></p>
                <p>Do you want to return to draft ?</p>
                """
                ),
            )
        },
    )
    def draft(self, by=None, description=None, **kwargs):
        if compliance_type := self.type:
            if by:
                self.changer = by.profile
            current_user = self.changer if self.changer else self.creator
            msg = _("{} has changed a Compliance Risk Review to Draft : <b>{}</b>").format(
                str(current_user), self.title
            )
            title = _("Compliance Risk Review : {}").format(self.title)
            recipients = ComplianceType.get_administrators(compliance_type).exclude(profile=current_user)
            self.notify(title, msg, recipients)

    @transition(
        field=status,
        source=Status.VALIDATION_REQUESTED,
        target=Status.VALIDATED,
        permission=can_active_request,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbcompliance:reviewcompliancetask",),
                icon=WBIcon.SEND.icon,
                key="validation",
                label="Validate",
                action_label=_("Validate"),
                description_fields=_(
                    """
                <p>Title: <b> {{title}} </b></p>
                <p>Status: <b>{{status}}</b></p> <p>From: <b>{{from_date}}</b></p>  <p>To: <b>{{to_date}}</b></p>
                <p>Do you want to validate?</p>
                """
                ),
            )
        },
    )
    def validation(self, by=None, description=None, **kwargs):
        if compliance_type := self.type:
            if by:
                self.changer = by.profile
            current_user = self.changer if self.changer else self.creator
            msg = _(
                """
                {} has validated a Compliance Risk Review : <b>{}</b>
            """
            ).format(str(current_user), self.title)
            title = _("Validation - Compliance Risk Review : {}").format(self.title)
            recipients = ComplianceType.get_administrators(compliance_type).exclude(profile=current_user)
            self.notify(title, msg, recipients)

    @transition(
        field=status,
        source=Status.VALIDATED,
        target=Status.DRAFT,
        permission=can_draft_request,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbcompliance:reviewcompliancetask",),
                icon=WBIcon.EDIT.icon,
                key="backtodraft",
                label="Back to draft",
                action_label=_("Back to draft"),
                description_fields=_(
                    """
                <p>Title: <b> {{title}} </b></p>
                <p>Status: <b>{{status}}</b></p> <p>From: <b>{{from_date}}</b></p>  <p>To: <b>{{to_date}}</b></p>
                <p>Do you want to return to the draft?</p>
                """
                ),
            )
        },
    )
    def backtodraft(self, by=None, description=None, **kwargs):
        if compliance_type := self.type:
            if by:
                self.changer = by.profile
            current_user = self.changer if self.changer else self.creator
            msg = _("{} has drafted a Compliance Indicators Report : <b>{}</b>").format(str(current_user), self.title)
            title = _("Compliance Indicators Report drafted: {}").format(self.title)
            recipients = ComplianceType.get_administrators(compliance_type)
            self.notify(title, msg, recipients)

    def compute_str(self) -> str:
        _str = "{}".format(self.title)
        if self.from_date or self.to_date:
            _str += " - ({} - {})".format(self.from_date, self.to_date)
        return _str

    def __str__(self) -> str:
        return self.computed_str

    def save(self, *args, **kwargs):
        self.computed_str = self.compute_str()
        super().save(*args, **kwargs)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcompliance:reviewcompliancetask"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcompliance:reviewcompliancetaskrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{computed_str}}"


@receiver(post_save, sender=ComplianceEvent)
def post_save_compliance_event(sender, instance, created, **kwargs):
    """
    Send notification to administrators
    """
    if created and (compliance_type := instance.type):
        current_profile = instance.changer if instance.changer else instance.creator
        recipients = ComplianceType.get_administrators(compliance_type).exclude(profile=current_profile)
        title = "{}: {}".format(ComplianceEvent.Type[instance.type_event].label, instance.title)
        msg = _("<p>An {} Event was created by {} {} at {}</p>").format(
            ComplianceEvent.Type[instance.type_event].label,
            current_profile.first_name,
            current_profile.last_name,
            instance.last_modified.strftime("%d-%b-%y %H:%M:%S"),
        )
        if instance.exec_summary:
            msg += _("<p> Summary : {}</p>").format(instance.exec_summary)
        for recipient in recipients:
            send_notification(
                code="wbcompliance.complianceevent.notify",
                title=title,
                body=msg,
                user=recipient,
                reverse_name="wbcompliance:complianceevent-detail",
                reverse_args=[instance.id],
            )



from dataclasses import dataclass
from datetime import date
from typing import TYPE_CHECKING, Any, Dict, Generator, Iterable, Optional

from django.contrib.contenttypes.models import ContentType
from django.db import models
from wbcore import serializers as wb_serializers

if TYPE_CHECKING:
    from wbcompliance.models.risk_management.incidents import RiskIncidentType


@dataclass(frozen=True)
class IncidentResult:
    breached_object: models.Model | None
    breached_object_repr: str
    breached_value: str | None
    report_details: dict[str, Any]
    severity: "RiskIncidentType"


class AbstractRuleBackend:
    OBJECT_FIELD_NAME = "evaluated_object"
    evaluated_object: Any
    evaluation_date: date
    parameters: dict[str, Any]

    def __init__(
        self,
        evaluation_date: date,
        evaluated_object: Any,
        json_parameters: Optional[Dict[str, Any]] = None,
        thresholds: Optional[
            Iterable
        ] = None,  # TODO refactor threshold to be a dataclass (DTO) to remove indirect dependency to the module
        **kwargs,
    ):
        if not json_parameters:
            json_parameters = {}
        self.evaluation_date = evaluation_date
        setattr(self, self.OBJECT_FIELD_NAME, evaluated_object)
        self.thresholds = thresholds if thresholds else []
        self.parameters = self._deserialize_parameters(json_parameters)

        for k, v in self.parameters.items():
            setattr(self, k, v)

    @classmethod
    def get_all_active_relationships(cls) -> Iterable:
        raise NotImplementedError()

    @classmethod
    def get_serializer_class(cls) -> wb_serializers.Serializer:
        """
        Return the serializer to deserialize the parameters given as json
        Returns:
            A serializer class
        Raises:
            NotImplementedError
        """
        raise NotImplementedError()

    @classmethod
    def _deserialize_parameters(cls, json_parameters) -> Dict[str, Any]:
        """
        The parameters are stored in the rule as a json field.

        E.g. As such, model are stored with their id.

        This allows for the implementing backend to override the default deserialization behavior
        Args:
            json_parameters: The serialized dictionary

        Returns:
            The deserializer dictionary. Default to returning the json parameters untouched
        """
        # If parameters are not empty, we expect the backend to define `get_serializer_class. This will crash otherwise
        serializer = cls.get_serializer_class()(data=json_parameters)
        if serializer.is_valid():
            return serializer.validated_data
        else:
            raise ValueError(serializer.errors)

    @classmethod
    def get_allowed_content_type(cls) -> "ContentType":
        """
        This function is called upon backend registration in order to determine the allowed checked_object content type

        Returns:
            The allowed content type

        Raises:
            NotImplementedError
        """
        raise NotImplementedError()

    def _build_dto_args(self):
        """
        Can be overrided to define what the default DataTransferObject is going to be
        It defaults to calling _build_dto on the evaluated object
        Returns:
            An object holding the DTO representation
        """
        if hasattr(self.evaluated_object, "_build_dto"):
            return self.evaluated_object._build_dto(self.evaluation_date)
        return tuple()

    def _process_dto(self, *dtos) -> Generator[IncidentResult, None, None]:
        """
        Check the rule against the instantiated backend given the DTO

        Returns:
            yield the breach object, its string representation, the incident report and the severity as IncidentResult

        Raises:
            NotImplementedError
        """
        raise NotImplementedError()

    def check_rule(self, *dto_args, **kwargs) -> Generator[IncidentResult, None, None]:
        """
        Build and Check the rule against the instantiated backend given its attributes

        Returns:
            yield the breach object, its string representation, the incident report and the severity as IncidentResult
        """
        if not dto_args:
            # We build the data transfer object if it is not provided.
            dto_args = self._build_dto_args()
        if any(dto_args):
            yield from self._process_dto(*dto_args, **kwargs)

    def is_passive_evaluation_valid(self) -> bool:
        """
        Determine if the instantiated backend can be evaluated given its attributes

        Returns:
            True if the backend rule can be evaluated
        """
        return True



from datetime import timedelta
from unittest.mock import patch

import pytest
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from faker import Faker
from guardian.shortcuts import assign_perm
from guardian.utils import get_anonymous_user
from pandas.tseries.offsets import BDay

from wbcompliance.models.risk_management.incidents import (
    RiskIncident,
    resolve_all_incidents_as_task,
)
from wbcompliance.models.risk_management.rules import RiskRule

User = get_user_model()

fake = Faker()


@pytest.mark.django_db
class TestRiskIncidentFactory:
    def test_init(self, risk_incident):
        assert risk_incident.computed_str

    def test_closing_incident(self, risk_incident, checked_object_incident_relationship_factory):
        checked_object_incident_relationship_factory.create(incident=risk_incident)
        assert risk_incident.checked_object_relationships.filter(status=RiskIncident.Status.OPEN).exists()
        risk_incident.status = RiskIncident.Status.CLOSED
        risk_incident.save()
        assert not risk_incident.checked_object_relationships.filter(status=RiskIncident.Status.OPEN).exists()

    def test_incident_date_range(
        self,
        risk_incident,
        checked_object_incident_relationship_factory,
        risk_check_factory,
        rule_checked_object_relationship_factory,
    ):
        rel1 = checked_object_incident_relationship_factory.create(incident=risk_incident)
        checked_object_relationship = rule_checked_object_relationship_factory.create(rule=risk_incident.rule)
        for i in range(7):
            next_date = (rel1.incident_date + BDay(i)).date()
            check = risk_check_factory.create(
                evaluation_date=next_date, rule_checked_object_relationship=checked_object_relationship
            )
            checked_object_incident_relationship_factory.create(incident=risk_incident, rule_check=check)
            risk_incident.refresh_from_db()
            assert risk_incident.date_range.upper == next_date + timedelta(days=1)
            assert RiskIncident.objects.count() == 1

    def test_incident_always_max_severity(
        self,
        risk_incident_factory,
        checked_object_incident_relationship_factory,
        risk_incident_type_factory,
    ):
        minor_severity = risk_incident_type_factory.create(severity_order=0)
        major_severity = risk_incident_type_factory.create(severity_order=1)
        critical_severity = risk_incident_type_factory.create(severity_order=2)

        incident = risk_incident_factory.create(severity=minor_severity)
        assert incident.severity == minor_severity

        checked_object_incident_relationship_factory.create(incident=incident, severity=minor_severity)
        incident.save()
        assert incident.severity == minor_severity

        checked_object_incident_relationship_factory.create(incident=incident, severity=major_severity)
        incident.save()
        assert incident.severity == major_severity

        incident.severity = critical_severity
        incident.save()
        assert incident.severity == critical_severity

    @pytest.mark.parametrize(
        "evaluation_date, is_breached_object",
        [
            (fake.date_object(), True),
            (fake.date_object(), False),
        ],
    )
    def test_get_or_create_incident(self, risk_rule, risk_incident_type, evaluation_date, is_breached_object):
        if is_breached_object:
            breached_object = get_anonymous_user()
            breached_object_repr = str(get_anonymous_user())
        else:
            breached_object = None
            breached_object_repr = fake.text(max_nb_chars=125)

        assert not RiskIncident.objects.exists()
        risk_rule.get_or_create_incident(evaluation_date, risk_incident_type, breached_object, breached_object_repr)
        if breached_object:
            assert (
                RiskIncident.objects.filter(
                    rule=risk_rule,
                    date_range__contains=evaluation_date,
                    breached_content_type=ContentType.objects.get_for_model(breached_object),
                    breached_object_id=breached_object.id,
                ).count()
                == 1
            )
        else:
            assert (
                RiskIncident.objects.filter(
                    rule=risk_rule, date_range__contains=evaluation_date, breached_object_repr=breached_object_repr
                ).count()
                == 1
            )
        assert RiskIncident.objects.count() == 1
        # Check if we reevaluate this incident one business day later, the incident is consider the same
        risk_rule.get_or_create_incident(
            (evaluation_date + BDay(1)).date(), risk_incident_type, breached_object, breached_object_repr
        )
        assert RiskIncident.objects.count() == 1

        # Check if we reevaluate this incident three business day later, the incident is consider the discontinue, and then a new incident is created
        risk_rule.get_or_create_incident(
            (evaluation_date + BDay(3)).date(), risk_incident_type, breached_object, breached_object_repr
        )

        assert RiskIncident.objects.count() == 2

    def test_get_or_create_incident_with_resolve_incident(
        self, risk_incident, checked_object_incident_relationship_factory
    ):
        subincident = checked_object_incident_relationship_factory.create(incident=risk_incident)
        assert risk_incident.status == RiskIncident.Status.OPEN
        risk_incident.resolve()
        risk_incident.save()
        assert risk_incident.status == RiskIncident.Status.RESOLVED
        new_evaluation_date = subincident.rule_check.evaluation_date + timedelta(days=1)
        incident, created = risk_incident.rule.get_or_create_incident(
            new_evaluation_date,
            subincident.severity,
            risk_incident.breached_content_object,
            risk_incident.breached_object_repr,
        )
        assert incident == risk_incident
        assert incident.date_range.upper == new_evaluation_date

    @pytest.mark.parametrize("incident_report", [fake.paragraph()])
    def test_update_or_create_relationship(
        self, risk_incident_factory, risk_check, incident_report, risk_incident_type
    ):
        risk_incident = risk_incident_factory.create(rule=risk_check.rule)
        assert not risk_incident.checked_object_relationships.exists()
        risk_incident.update_or_create_relationship(
            risk_check, incident_report, dict(), fake.pyfloat(), risk_incident_type
        )
        assert risk_incident.checked_object_relationships.filter(
            rule_check=risk_check, severity=risk_incident_type
        ).exists()

    @pytest.mark.parametrize(
        "incident_report, override_incident",
        [
            (fake.paragraph(), True),
            (fake.paragraph(), False),
        ],
    )
    def test_update_or_create_relationship_already_existing(
        self, risk_incident_factory, risk_check_factory, incident_report, risk_incident_type, override_incident
    ):
        check1 = risk_check_factory.create()
        check2 = risk_check_factory.create(
            rule=check1.rule,
            checked_object_id=check1.checked_object_id,
            checked_object_content_type=check1.checked_object_content_type,
            evaluation_date=check1.evaluation_date,
        )
        risk_incident = risk_incident_factory.create(rule=check1.rule)

        # We check the first time
        initial_breached_value = str(fake.pyfloat())
        risk_incident.update_or_create_relationship(
            check1, incident_report, dict(), initial_breached_value, risk_incident_type
        )
        # verify that the relationship was created
        assert risk_incident.checked_object_relationships.count() == 1
        assert (
            risk_incident.checked_object_relationships.filter(rule_check=check1, severity=risk_incident_type).count()
            == 1
        )

        # Closing incident because open incident are override by default
        risk_incident.status = RiskIncident.Status.CLOSED
        risk_incident.save()
        breached_value = str(fake.pyfloat())
        risk_incident.update_or_create_relationship(
            check2, incident_report, dict(), breached_value, risk_incident_type, override_incident
        )
        subincidents = risk_incident.checked_object_relationships.all()
        if not override_incident:
            assert subincidents.get(rule_check=check1).breached_value == initial_breached_value
            assert subincidents.get(rule_check=check2).breached_value == breached_value
        else:
            assert subincidents.get(rule_check=check1).breached_value == breached_value
            assert not subincidents.filter(rule_check=check2).exists()
        # asser the parent incident were reopened
        risk_incident.refresh_from_db()
        if not override_incident:
            assert risk_incident.status == RiskIncident.Status.OPEN
        else:
            assert risk_incident.status == RiskIncident.Status.CLOSED

        if not override_incident:
            risk_incident.status = RiskIncident.Status.CLOSED
            risk_incident.save()
            # we submit a new check the next day with the same breached value. We test if the parent incident remains close because of it
            check3 = risk_check_factory.create(
                rule=check1.rule,
                checked_object_id=check1.checked_object_id,
                checked_object_content_type=check1.checked_object_content_type,
                evaluation_date=check1.evaluation_date + timedelta(days=1),
            )
            risk_incident.update_or_create_relationship(
                check3, incident_report, dict(), breached_value, risk_incident_type
            )
            risk_incident.refresh_from_db()
            assert risk_incident.status == RiskIncident.Status.CLOSED
            assert (
                subincidents.get(rule_check=check3, breached_value=breached_value).status == RiskIncident.Status.CLOSED
            )

    @patch.object(RiskIncident, "resolve_all_incidents")
    @pytest.mark.parametrize("reviewer_comment", [(fake.paragraph())])
    def test_resolve_all_incidents_as_task(self, mock_function, risk_incident, user, reviewer_comment):
        resolve_all_incidents_as_task(user.id, reviewer_comment, risk_incident.id)
        assert mock_function.call_count == 1

    @pytest.mark.parametrize(
        "is_superuser, is_admin",
        [
            (True, True),
            (False, True),
            (False, False),
            (True, False),
            (False, False),
        ],
    )
    def test_can_manage_without_incident(self, user, is_superuser, is_admin):
        if is_superuser:
            user.is_superuser = True
            user.save()
        if is_admin:
            admin_perm = Permission.objects.get(codename="administrate_riskrule")
            user.user_permissions.add(admin_perm)
        res = RiskIncident.can_manage(user)
        assert res == (is_admin or is_superuser)

    def test_can_manage_with_incident(self, user, risk_incident):
        assert not RiskIncident.can_manage(user, risk_incident.rule)
        assign_perm(RiskRule.change_perm_str, user, risk_incident.rule)
        assert RiskIncident.can_manage(user, risk_incident.rule)

    @pytest.mark.parametrize("risk_rule__automatically_close_incident", [True, False])
    def test_post_workflow_change_status(self, risk_incident_factory, risk_rule):
        risk_incident = risk_incident_factory.create(rule=risk_rule)
        assert not risk_incident.status == RiskIncident.Status.CLOSED
        risk_incident.post_workflow()
        assert (
            risk_incident.status == RiskIncident.Status.CLOSED
            if risk_incident.rule.automatically_close_incident
            else RiskIncident.Status.OPEN
        )

    def test_post_workflow_elevate_incident(self, risk_incident, rule_threshold_factory):
        current_threshold = risk_incident.threshold
        current_threshold.upgradable_after_days = risk_incident.business_days - 1
        current_threshold.save()
        next_threshold_1 = rule_threshold_factory.create(
            rule=risk_incident.rule, upgradable_after_days=risk_incident.business_days - 1
        )
        assert next_threshold_1.severity.severity_order == 1
        next_threshold_2 = rule_threshold_factory.create(rule=risk_incident.rule)
        assert next_threshold_2.severity.severity_order == 2
        assert risk_incident.threshold == current_threshold
        risk_incident.post_workflow()
        assert risk_incident.threshold == next_threshold_1
        risk_incident.post_workflow()
        assert risk_incident.threshold == next_threshold_2
        risk_incident.post_workflow()
        assert risk_incident.threshold == next_threshold_2


@pytest.mark.django_db
class TestCheckedObjectIncidentRelationshipFactory:
    def test_init(self, rule_checked_object_relationship):
        assert rule_checked_object_relationship

    def test_init_checked_object_not_valid(
        self,
        checked_object_incident_relationship_factory,
        rule_checked_object_relationship_factory,
        risk_incident_factory,
        risk_check_factory,
    ):
        rel_user = rule_checked_object_relationship_factory.create()
        rel_other = rule_checked_object_relationship_factory.create(
            checked_object_content_type=ContentType.objects.get_for_model(rel_user)
        )

        incident = risk_incident_factory.create(rule=rel_user.rule)
        check = risk_check_factory.create(rule_checked_object_relationship=rel_other)
        with pytest.raises(ValidationError):
            checked_rel = checked_object_incident_relationship_factory.create(incident=incident, rule_check=check)
            checked_rel.full_clean()

    def test_init_backend_allowed_content_type_not_valid(
        self,
        checked_object_incident_relationship_factory,
        rule_checked_object_relationship_factory,
        risk_incident_factory,
    ):
        rel_user = rule_checked_object_relationship_factory.create()
        rel_user.rule.rule_backend.allowed_checked_object_content_type = ContentType.objects.get_for_model(rel_user)
        rel_user.rule.rule_backend.save()
        incident = risk_incident_factory.create(rule=rel_user.rule)
        with pytest.raises(ValidationError):
            checked_rel = checked_object_incident_relationship_factory.create(incident=incident)
            checked_rel.full_clean()

    def test_checked_object(self, checked_object_incident_relationship):
        assert checked_object_incident_relationship.checked_object
        assert (
            checked_object_incident_relationship.checked_object
            == checked_object_incident_relationship.rule_check.checked_object
        )



from datetime import date, timedelta
from typing import TYPE_CHECKING, Any, Optional

import pandas as pd
from celery import shared_task
from django.contrib.auth import get_user_model
from django.contrib.auth.models import User as BaseUser
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.fields import DateRangeField
from django.core.exceptions import ValidationError
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.db.models import Value
from django.db.models.functions import TruncDate
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils.translation import gettext as _
from django_fsm import FSMField, transition
from guardian.core import ObjectPermissionChecker
from psycopg.types.range import DateRange
from wbcore.content_type.utils import get_ancestors_content_type
from wbcore.contrib.color.enums import WBColor
from wbcore.contrib.directory.models import Person
from wbcore.contrib.icons import WBIcon
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.enums import RequestType
from wbcore.metadata.configs.buttons import ActionButton, ButtonDefaultColor
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)
from wbcore.models import WBModel
from wbcore.utils.models import ComplexToStringMixin

User: BaseUser = get_user_model()

if TYPE_CHECKING:
    from wbcompliance.models import RiskCheck


class RiskIncidentType(models.Model):
    name = models.CharField(max_length=128)
    severity_order = models.PositiveIntegerField(default=0, unique=True)

    color = models.CharField(max_length=20, verbose_name=_("Color"), default=WBColor.YELLOW_LIGHT.value)
    is_ignorable = models.BooleanField(default=True, verbose_name=_("Can be ignored"))
    is_automatically_closed = models.BooleanField(default=False, verbose_name=_("Automatically closed"))
    is_informational = models.BooleanField(
        default=False,
        verbose_name=_("Only Informational"),
        help_text=_("If true, the associated rule is not considered an incident"),
    )

    def __str__(self) -> str:
        return self.name

    def save(self, *args, **kwargs):
        if (
            self.pk
            and RiskIncidentType.objects.exclude(id=self.pk).filter(severity_order=self.severity_order).exists()
        ):
            self.severity_order += 1
        return super().save(*args, **kwargs)

    class Meta:
        ordering = ("severity_order",)
        verbose_name = "Risk Incident Type"
        verbose_name_plural = "Risk Incidents Type"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcompliance:riskincidenttyperepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{name}}"


class RiskIncidentMixin(models.Model):
    class Status(models.TextChoices):
        OPEN = "OPEN", "Open"  # Newly created incidents
        RESOLVED = "RESOLVED", "Resolved"
        IGNORED = "IGNORED", "Ignored"
        CLOSED = "CLOSED", "Closed"

    severity = models.ForeignKey(
        "wbcompliance.RiskIncidentType",
        on_delete=models.CASCADE,
        verbose_name=_("Severity"),
        related_name="%(class)s",
    )
    comment = models.TextField(blank=True, null=True, verbose_name=_("Comment"))

    resolved_by = models.ForeignKey(
        "directory.Person",
        on_delete=models.SET_NULL,
        verbose_name=_("Handled by"),
        help_text=_("The person that resolved or ignored this incident"),
        related_name="%(class)s_handled",
        blank=True,
        null=True,
    )

    class Meta:
        abstract = True


class CheckedObjectIncidentRelationship(ComplexToStringMixin, RiskIncidentMixin, WBModel):
    incident = models.ForeignKey(
        to="wbcompliance.RiskIncident",
        related_name="checked_object_relationships",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )

    rule_check = models.ForeignKey(
        "wbcompliance.RiskCheck",
        on_delete=models.CASCADE,
        verbose_name=_("Check"),
        help_text=_("The check that opened this incident"),
        related_name="incidents",
    )
    breached_value = models.CharField(
        blank=True,
        null=True,
        verbose_name=_("Breached Value"),
        max_length=128,
        help_text="The value that breached the rule threshold, can be None",
    )

    report = models.TextField(blank=True, null=True, verbose_name=_("Report"))
    report_details = models.JSONField(default=dict, blank=True, encoder=DjangoJSONEncoder)

    # A user can mark an incident as resolved, if the necessary actions were taken
    status = FSMField(
        default=RiskIncidentMixin.Status.OPEN, choices=RiskIncidentMixin.Status.choices, verbose_name=_("Status")
    )

    @transition(
        status,
        [RiskIncidentMixin.Status.OPEN],
        RiskIncidentMixin.Status.RESOLVED,
        permission=lambda instance, user: RiskIncident.can_manage(user, instance.rule),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbcompliance:riskincident",),
                icon=WBIcon.APPROVE.icon,
                color=ButtonDefaultColor.SUCCESS,
                key="resolve",
                label=_("Resolve"),
                action_label=_("Resolve"),
                description_fields=_("<p>Are you sure you want to resolve the incident {{computed_str}}"),
                instance_display=create_simple_display([["comment"]]),
            )
        },
    )
    def resolve(self, by=None, **kwargs):
        if by:
            self.resolved_by = by.profile

    @transition(
        status,
        [RiskIncidentMixin.Status.OPEN],
        RiskIncidentMixin.Status.IGNORED,
        permission=lambda instance, user: RiskIncident.can_manage(user, instance.rule),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbcompliance:riskincident",),
                icon=WBIcon.IGNORE.icon,
                color=ButtonDefaultColor.WARNING,
                key="ignore",
                label=_("Ignore"),
                action_label=_("Ignore"),
                description_fields=_("<p>Are you sure you want to ignore the incident {{computed_str}}"),
                instance_display=create_simple_display([["severity"], ["comment"]]),
            )
        },
    )
    def ignore(self, by=None, **kwargs):
        if by:
            self.resolved_by = by.profile

    def can_ignore(self):
        if not self.severity.is_ignorable:
            return {"severity": [_("Incident type {} is not ignorable").format(self.severity)]}
        return dict()

    @property
    def incident_date(self) -> date:
        return self.rule_check.evaluation_date

    @property
    def rule(self) -> Any:
        if self.incident:
            return self.incident.rule
        return self.rule_check.rule

    @property
    def checked_object(self) -> models.Model:
        return self.rule_check.checked_object

    def clean(self):
        allowed_checked_object_content_type = self.rule.rule_backend.allowed_checked_object_content_type
        if not self.checked_object:
            raise ValidationError("Checked Object cannot be null")
        checked_object_content_type = ContentType.objects.get_for_model(self.checked_object)
        if allowed_checked_object_content_type and (
            allowed_checked_object_content_type not in list(get_ancestors_content_type(checked_object_content_type))
        ):
            raise ValidationError(
                _(
                    "The relationship content type ({}) needs to match the incident rule backend allowed content type ({})"
                ).format(self.checked_object, allowed_checked_object_content_type)
            )
        super().clean()

    class Meta:
        verbose_name = "Incident to Checked Object relationship"
        verbose_name_plural = "Incident to Checked Object relationships"
        unique_together = ("incident", "rule_check")
        indexes = [
            models.Index(fields=["incident", "rule_check"]),
        ]

    def compute_str(self) -> str:
        return _("{} {} Sub Incident for checked_object {}").format(self.status, self.severity, self.rule_check)

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcompliance:checkedobjectincidentrelationshiprepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcompliance:checkedobjectincidentrelationship"


class RiskIncidentDefaultManager(models.Manager):
    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .annotate(
                ignore_until_time=models.F("last_ignored_date")
                + models.F("ignore_duration")
                + Value(timedelta(days=1)),
                ignore_until=TruncDate("ignore_until_time"),
            )
        )


class RiskIncident(ComplexToStringMixin, RiskIncidentMixin, WBModel):
    """
    Instance defining the incident that has happened during a check initiated by a certain rule
    """

    date_range = DateRangeField(blank=True, null=True, help_text=_("The incident spans date interval"))
    rule = models.ForeignKey(
        "wbcompliance.RiskRule",
        on_delete=models.CASCADE,
        verbose_name=_("Rule"),
        help_text=_("The rule that opened this incident"),
        related_name="incidents",
    )

    breached_content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE, null=True, blank=True)
    breached_object_id = models.PositiveIntegerField(null=True, blank=True)
    breached_content_object = GenericForeignKey("breached_content_type", "breached_object_id")
    breached_object_repr = models.CharField(
        max_length=128,
        blank=True,
        null=True,
        verbose_name=_("Breached Object Representation"),
        help_text=_("String Representation of the breached object"),
    )

    status = FSMField(
        default=RiskIncidentMixin.Status.OPEN, choices=RiskIncidentMixin.Status.choices, verbose_name=_("Status")
    )
    is_notified = models.BooleanField(
        default=False, verbose_name=_("Notified"), help_text=_("True if the incident is already notified to the users")
    )

    last_ignored_date = models.DateField(blank=True, null=True)
    ignore_duration = models.DurationField(
        blank=True, null=True, help_text=_("If set, will ignore the forthcoming incidents for the specified duration")
    )

    def get_ignore_until_date(self) -> date | None:
        if self.last_ignored_date and self.ignore_duration:
            return self.last_ignored_date + self.ignore_duration

    @transition(
        status,
        [RiskIncidentMixin.Status.OPEN],
        RiskIncidentMixin.Status.RESOLVED,
        permission=lambda instance, user: RiskIncident.can_manage(user, instance.rule),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbcompliance:riskincident",),
                icon=WBIcon.APPROVE.icon,
                color=ButtonDefaultColor.SUCCESS,
                key="resolve",
                label=_("Resolve"),
                action_label=_("Resolve"),
                description_fields=_(
                    "<p>Are you sure you want to resolve the incident {{computed_str}} and its relationships?"
                ),
                instance_display=create_simple_display([["comment"]]),
            )
        },
    )
    def resolve(self, by=None, **kwargs):
        if by:
            self.resolved_by = by.profile
            self.checked_object_relationships.filter(status=self.Status.OPEN).update(
                status=self.status, resolved_by=self.resolved_by
            )

    @transition(
        status,
        [RiskIncidentMixin.Status.OPEN],
        RiskIncidentMixin.Status.IGNORED,
        permission=lambda instance, user: RiskIncident.can_manage(user, instance.rule),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbcompliance:riskincident",),
                icon=WBIcon.IGNORE.icon,
                color=ButtonDefaultColor.WARNING,
                key="ignore",
                label=_("Ignore"),
                action_label=_("Ignore"),
                description_fields=_(
                    "<p>Are you sure you want to ignore the incident {{computed_str}} and its relationships? If you set a days greater than 0, the forthcoming incident will be automatically ignore"
                ),
                instance_display=create_simple_display(
                    [["severity", "ignore_duration_in_days"], ["comment", "comment"]]
                ),
            )
        },
    )
    def ignore(self, by=None, **kwargs):
        if by:
            self.last_ignored_date = self.date_range.upper  # type: ignore
            self.resolved_by = by.profile
            self.checked_object_relationships.filter(status=self.Status.OPEN).update(
                status=self.status, resolved_by=self.resolved_by
            )

    def can_ignore(self):
        if not self.severity.is_ignorable:
            return {"severity": [_("Incident type {} is not ignorable").format(self.severity)]}
        return dict()

    objects = RiskIncidentDefaultManager()

    class Meta:
        verbose_name = "Risk Incident"
        verbose_name_plural = "Risk Incidents"

        notification_types = [
            create_notification_type(
                code="wbcompliance.riskincident.notify",
                title="Risk Incident Notification",
                help_text="Notifies you when an incident is triggered.",
                email=True,
                web=False,
                mobile=False,
                is_lock=True,
            )
        ]

    def save(self, *args, **kwargs):
        if not self.breached_object_repr:
            self.breached_object_repr = str(self.breached_content_object)

        if self.id and self.checked_object_relationships.exists():
            # If this global incident is closed, we close all opened incident relationships
            if self.status != self.Status.OPEN:
                self.checked_object_relationships.filter(status=self.Status.OPEN).update(status=self.status)

            existing_severity_orders = list(
                self.checked_object_relationships.values_list("severity__severity_order", flat=True)
            )
            self.severity = RiskIncidentType.objects.get(
                severity_order=max([self.severity.severity_order, *existing_severity_orders])
            )
            self.date_range = DateRange(
                lower=self.checked_object_relationships.earliest("rule_check__evaluation_date").incident_date,
                upper=self.checked_object_relationships.latest("rule_check__evaluation_date").incident_date
                + timedelta(days=1),
            )  # type: ignore
        super().save(*args, **kwargs)

    @property
    def date_range_lower_repr(self) -> str:
        return f"[{self.date_range.lower:%Y-%m-%d}" if self.date_range.lower else "]-∞"  # type: ignore

    @property
    def date_range_upper_repr(self) -> str:
        return f"{self.date_range.upper - timedelta(days=1):%Y-%m-%d}]" if self.date_range.upper else "+∞["  # type: ignore

    @property
    def automatically_close_incident(self) -> bool:
        return self.rule.automatically_close_incident or self.severity.is_automatically_closed

    def compute_str(self) -> str:
        if self.severity:
            return _("({}) {} Incident for breached object {} during {},{}").format(
                self.status,
                self.severity,
                self.breached_content_object,
                self.date_range_lower_repr,
                self.date_range_upper_repr,
            )
        return _("({}) Incident for breached object {} during {}").format(
            self.status, self.breached_content_object, self.date_range_lower_repr, self.date_range_upper_repr
        )

    def update_or_create_relationship(
        self,
        check: "RiskCheck",
        incident_report: str,
        incident_report_details: dict[str, Any],
        breached_value: str | None,
        incident_severity: "RiskIncidentType",
        override_incident: bool | None = False,
    ):
        # We assume that if the incident was not created and its status is other than opened, then it is a recheck and this incident was already handled.
        checked_object = check.checked_object
        similar_incidents_relationships = CheckedObjectIncidentRelationship.objects.filter(
            incident=self,
            rule_check__checked_object_content_type=ContentType.objects.get_for_model(checked_object),
            rule_check__checked_object_id=checked_object.id,
        ).distinct()
        defaults = {
            "report": incident_report,
            "report_details": incident_report_details,
            "breached_value": breached_value,
            "severity": incident_severity,
        }
        if self.rule.automatically_close_incident:
            defaults["status"] = RiskIncident.Status.CLOSED
        # if a previous check created an incident with the same breached value, we don't reopen it
        if override_incident or (
            (previous_check := check.previous_check)
            and similar_incidents_relationships.filter(rule_check=previous_check, breached_value=breached_value)
        ):
            defaults["status"] = self.status

        potential_existing_incidents_relationships = similar_incidents_relationships.filter(
            rule_check__evaluation_date=check.evaluation_date
        )
        if not potential_existing_incidents_relationships.exists() or not override_incident:
            rel, _ = CheckedObjectIncidentRelationship.objects.update_or_create(
                incident=self, rule_check=check, defaults=defaults
            )
            rel.full_clean()

        else:
            potential_existing_incidents_relationships.update(**defaults)

    @property
    def threshold(self) -> Any:
        return self.rule.thresholds.get(severity=self.severity)

    @property
    def notifiable(self) -> bool:
        """
        Property for wether the incident can fire notification
        Returns:
            True if the incident can fire notification
        """
        return not self.rule.is_silent and self.threshold.get_notifiable_users().exists()

    @property
    def business_days(self) -> int:
        """
        Property to get the number of days this incident span

        Returns:
            The number of business days
        """
        return len(pd.date_range(self.date_range.lower, self.date_range.upper, freq="B", inclusive="left"))  # type: ignore

    def post_workflow(self):
        """
        Post save method to check the extra mechanisms

        """

        # Check if the incident is automatically closed
        if self.automatically_close_incident:
            self.status = RiskIncident.Status.CLOSED
        # Check if the incident needs to be elevated
        if (
            (upgradable_after_days := self.threshold.upgradable_after_days)
            and (next_threshold := self.threshold.next_threshold)
            and (self.business_days > upgradable_after_days)
        ):
            self.severity = next_threshold.severity

        self.save()

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcompliance:riskincident"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcompliance:riskincidentrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def resolve_all_incidents(
        cls,
        resolved_by: "Person",
        reviewer_comment: str,
        is_resolved: Optional[bool] = True,
        rule_id: Optional[int] = None,
    ):
        """
        Utility methods to close all incidents that were created during an optional check and/or triggered by an optional rule
        Args:
            resolved_by: The user (Person) resolving all incidents.
            reviewer_comment: The resolver's comment.
            is_resolved: True if incident state becomes "RESOLVED", "IGNORED" otherwise. Defaults to True.
            risk_check_id: Optional check id. Defaults to None (Closing all incidents from any check).
            rule_id: Optional Rule id. Defaults to None (Closing all incidents from any rule).
        """
        qs = RiskIncident.objects.filter(status=RiskIncident.Status.OPEN)
        if rule_id:
            qs = qs.filter(rule__id=rule_id)
        for incident in qs.iterator():
            if is_resolved:
                incident.resolve(by=resolved_by)
            else:
                incident.ignore(by=resolved_by)
            incident.comment = reviewer_comment
            incident.save()

    @classmethod
    def can_manage(cls, user: "User", rule: Optional[models.Model] = None) -> bool:
        """
        Utility function to check wether the given user can manage|edit all incidents (manager) or on optional incident
        Args:
            user: The user whose permission is checked.
            incident: Optional Incident. If none, assume global permission check. Defaults to None.

        Returns:
            True if user can manage
        """
        from .rules import RiskRule

        if rule:
            checker = ObjectPermissionChecker(user)
            return checker.has_perm(RiskRule.change_perm_str, rule) or checker.has_perm(RiskRule.admin_perm_str, rule)
        return user.has_perm(RiskRule.admin_perm_str)


@receiver(post_save, sender="wbcompliance.CheckedObjectIncidentRelationship")
def post_save_incident_relationship(sender, instance, created, raw, **kwargs):
    """
    Trigger notification on incident creation
    """
    if not raw and instance.incident:
        if (
            created
            and instance.status == RiskIncident.Status.OPEN
            and not instance.severity.is_informational  # do not reopen incident from informational severity
            and (
                instance.incident.status in [RiskIncident.Status.RESOLVED, RiskIncident.Status.CLOSED]
                or (
                    instance.incident.status == RiskIncident.Status.IGNORED
                    and (
                        not (ignore_until := instance.incident.get_ignore_until_date())
                        or ignore_until < instance.incident_date
                    )
                )
            )
        ):
            instance.incident.status = RiskIncident.Status.OPEN
            instance.incident.last_ignored_date = None
            instance.incident.ignore_duration = None

        instance.incident.save()


@shared_task
def resolve_all_incidents_as_task(
    resolved_by_id,
    reviewer_comment,
    is_resolved: Optional[bool] = True,
    rule_id: Optional[int] = None,
):
    """
    Async task to resolve all incidents
    """
    resolved_by = User.objects.get(id=resolved_by_id)
    RiskIncident.resolve_all_incidents(resolved_by, reviewer_comment, is_resolved, rule_id=rule_id)



from django.utils.module_loading import autodiscover_modules
from django.db import DEFAULT_DB_ALIAS
from django.apps import apps as global_apps


def autodiscover_riskmanagement_backends(
    app_config, verbosity=2, interactive=True, using=DEFAULT_DB_ALIAS, apps=global_apps, **kwargs
):
    # we wrap the autodiscover into a post_migrate receiver because we expect db calls
    autodiscover_modules("risk_management.backends")



from datetime import timedelta
from typing import Optional

from django.utils.translation import gettext as _
from wbcore.enums import Unit
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

from wbcompliance.models import (
    ComplianceAction,
    ComplianceEvent,
    ComplianceTaskGroup,
    ComplianceTaskInstance,
    ReviewComplianceTask,
)
from wbcompliance.models.enums import IncidentSeverity


def get_legends(model=None):
    """
    Dynamically create the activity legend based on Activity Enum
    """
    list_format = []
    if model:
        for _status, color in model.Status.get_color_map():
            list_format.append(dp.LegendItem(icon=color, label=_status.label, value=_status.value))
    return [dp.Legend(key="status", items=list_format)]


def get_list_formatting(model=None):
    """
    Dynamically create the activity list formatting based on Activity Enum
    """
    color_conditions = []
    if model:
        for _status, color in model.Status.get_color_map():
            color_conditions.append(
                dp.FormattingRule(condition=("==", _status.name), style={"backgroundColor": color})
            )
    return [
        dp.Formatting(column="status", formatting_rules=color_conditions),
    ]


class ComplianceTaskGroupDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="name", label=_("Name")),
                dp.Field(key="order", label=_("Order")),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [["name", "order"], [repeat_field(2, "compliancetask_section")]],
            [
                create_simple_section(
                    "compliancetask_section", _("Indicators"), [["compliancetask"]], "compliancetask", collapsed=False
                )
            ],
        )


class ComplianceTaskDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        color_conditions = []
        list_format = []
        for _type, color in IncidentSeverity.get_color_map():
            color_conditions.append(dp.FormattingRule(condition=("==", _type.name), style={"backgroundColor": color}))
            list_format.append(dp.LegendItem(icon=color, label=_type.label, value=_type.value))

        return dp.ListDisplay(
            fields=[
                dp.Field(key="title", label=_("Title"), width=Unit.PIXEL(600)),
                dp.Field(key="occurrence", label=_("Occurrence")),
                dp.Field(key="risk_level", label=_("Risk Level")),
                dp.Field(key="active", label=_("Active")),
                dp.Field(key="group", label=_("Group"), width=Unit.PIXEL(350)),
                dp.Field(key="type", label=_("Administrator")),
                dp.Field(key="review", label=_("Indicator Reports"), width=Unit.PIXEL(300)),
            ],
            legends=[dp.Legend(key="risk_level", items=list_format)],
            formatting=[
                dp.Formatting(column="risk_level", formatting_rules=color_conditions),
            ],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["title", "title", "type"],
                ["occurrence", "risk_level", "active"],
                ["review", "review", "group"],
                [repeat_field(3, "description")],
                [repeat_field(3, "remarks")],
                [repeat_field(3, "compliancetaskinstance_section")],
            ],
            [
                create_simple_section(
                    "compliancetaskinstance_section",
                    _("Instances"),
                    [["compliancetaskinstance"]],
                    "compliancetaskinstance",
                    collapsed=True,
                )
            ],
        )


class ComplianceTaskReviewDisplayConfig(ComplianceTaskDisplayConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        color_conditions = []
        list_format = []
        for _type, color in IncidentSeverity.get_color_map():
            color_conditions.append(dp.FormattingRule(condition=("==", _type.name), style={"backgroundColor": color}))
            list_format.append(dp.LegendItem(icon=color, label=_type.label, value=_type.value))

        return dp.ListDisplay(
            fields=[
                dp.Field(key="title", label=_("Title"), width=Unit.PIXEL(400)),
                dp.Field(key="description", label=_("Description"), width=Unit.PIXEL(400)),
                dp.Field(key="occurrence", label=_("Occurrence"), width=Unit.PIXEL(150)),
                dp.Field(key="risk_level", label=_("Risk level"), width=Unit.PIXEL(150)),
                dp.Field(key="remarks", label=_("Remarks"), width=Unit.PIXEL(250)),
                dp.Field(key="group", label=_("Group"), width=Unit.PIXEL(250)),
                dp.Field(key="type", label=_("Administrator"), width=Unit.PIXEL(160)),
            ],
            legends=[dp.Legend(key="risk_level", items=list_format)],
            formatting=[
                dp.Formatting(column="risk_level", formatting_rules=color_conditions),
            ],
        )


class ComplianceTaskInstanceDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="task", label=_("Indicators")),
                dp.Field(key="occured", label=_("Occured")),
                dp.Field(key="status", label=_("Status")),
                dp.Field(key="type_name", label=_("Administrator")),
                dp.Field(key="group_name", label=_("Group")),
                dp.Field(key="review", label=_("Indicator Instance Report")),
            ],
            legends=get_legends(ComplianceTaskInstance),
            formatting=get_list_formatting(ComplianceTaskInstance),
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                [repeat_field(2, "type_name")],
                ["task", "review"],
                ["status", "occured"],
                [repeat_field(2, "text")],
                [repeat_field(2, "summary_text")],
            ]
        )


class ComplianceTaskInstanceComplianceTaskDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="occured", label=_("Occured")),
                dp.Field(key="status", label=_("Status")),
                dp.Field(key="type_name", label=_("Administrator")),
                dp.Field(key="group_name", label=_("Group")),
                dp.Field(key="review", label=_("Indicator Instance Report")),
            ],
            legends=get_legends(ComplianceTaskInstance),
            formatting=get_list_formatting(ComplianceTaskInstance),
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                [repeat_field(2, "type_name")],
                [repeat_field(2, "review")],
                ["status", "occured"],
                [repeat_field(2, "text")],
                [repeat_field(2, "summary_text")],
            ]
        )


class ComplianceTaskInstanceReviewDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="occured", label=_("Occured"), width=Unit.PIXEL(90)),
                dp.Field(key="task_title", label=_("Indicators"), width=Unit.PIXEL(380)),
                dp.Field(key="status", label=_("Status"), width=Unit.PIXEL(200)),
                dp.Field(key="text", label=_("Text"), width=Unit.PIXEL(500)),
                dp.Field(key="summary_text", label=_("Summary Text"), width=Unit.PIXEL(400)),
                dp.Field(key="type_name", label=_("Administrator"), width=Unit.PIXEL(100)),
            ],
            legends=get_legends(ComplianceTaskInstance),
            formatting=get_list_formatting(ComplianceTaskInstance),
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                [repeat_field(2, "type_name")],
                [repeat_field(2, "review")],
                ["status", "occured"],
                [repeat_field(2, "text")],
                [repeat_field(2, "summary_text")],
            ]
        )


class ComplianceTaskMatrixPandasDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        _color_conditions = []
        for _status, color in ComplianceTaskInstance.Status.get_color_map():
            _color_conditions.append(
                dp.FormattingRule(condition=("==", _status.label), style={"backgroundColor": color})
            )

        _fields = [
            dp.Field(key="type_name", label=_("Administrator"), width=Unit.PIXEL(100)),
            dp.Field(key="group_name", label=_("Group"), width=Unit.PIXEL(200)),
            dp.Field(
                key="task_title",
                label=_("Indicator"),
                width=Unit.PIXEL(270),
                formatting_rules=[dp.FormattingRule(condition=("!=", ""), style={"font-weight": "bold"})],
            ),
        ]
        if _dict := ComplianceTaskInstance.get_dict_max_count_task():
            qs_occured = (
                ComplianceTaskInstance.objects.filter(task=_dict.get("task"))
                .order_by("-occured")
                .values_list("occured", flat=True)
                .distinct()[:12]
            )
            for date in qs_occured:
                date_str = date.strftime("%Y-%m-%d")
                covered_month = (date - timedelta(days=1)).strftime("%Y-%h")
                _fields.append(
                    dp.Field(
                        key=date_str, label=covered_month, formatting_rules=_color_conditions, width=Unit.PIXEL(150)
                    )
                )
        return dp.ListDisplay(fields=_fields)


class ComplianceActionDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="title", label=_("Title")),
                dp.Field(key="status", label=_("Status")),
                dp.Field(key="deadline", label=_("Deadline")),
                dp.Field(key="progress", label=_("Progress")),
                dp.Field(key="active", label=_("Active")),
                dp.Field(key="type", label="Administrator"),
                dp.Field(key="creator", label="Creator"),
                dp.Field(key="changer", label="Changer"),
                dp.Field(key="last_modified", label="Changed"),
            ],
            legends=get_legends(ComplianceAction),
            formatting=get_list_formatting(ComplianceAction),
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                [repeat_field(2, "title")],
                [repeat_field(2, "type")],
                ["active", "deadline"],
                ["status", "progress"],
                ["creator", "created"],
                ["changer", "last_modified"],
                [repeat_field(2, "description")],
                [repeat_field(2, "summary_description")],
            ]
        )


class ComplianceEventDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        color_conditions = []
        list_format = []
        for _type, color in ComplianceEvent.Type.get_color_map():
            color_conditions.append(dp.FormattingRule(condition=("==", _type.name), style={"backgroundColor": color}))
            list_format.append(dp.LegendItem(icon=color, label=_type.label, value=_type.value))

        return dp.ListDisplay(
            fields=[
                dp.Field(key="title", label="Title"),
                dp.Field(key="type_event", label="Type Event"),
                dp.Field(key="level", label="Level"),
                dp.Field(key="active", label="Active"),
                dp.Field(key="type", label="Administrator"),
                dp.Field(key="creator", label="Creator"),
                dp.Field(key="changer", label="Changer"),
                dp.Field(key="last_modified", label="Changed"),
                dp.Field(key="confidential", label="Confidential"),
            ],
            legends=[dp.Legend(key="type_event", items=list_format)],
            formatting=[
                dp.Formatting(column="type_event", formatting_rules=color_conditions),
            ],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                [repeat_field(2, "title")],
                [repeat_field(2, "type")],
                ["active", "confidential"],
                ["type_event", "level"],
                ["creator", "created"],
                ["changer", "last_modified"],
                [repeat_field(2, "exec_summary")],
                [repeat_field(2, "exec_summary_board")],
                [repeat_field(2, "description")],
                [repeat_field(2, "actions_taken")],
                [repeat_field(2, "consequences")],
                [repeat_field(2, "future_suggestions")],
            ]
        )


class ReviewComplianceTaskDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="year", label="Year"),
                dp.Field(key="occured", label="Occured Instance"),
                dp.Field(key="title", label="Title"),
                dp.Field(key="from_date", label="From"),
                dp.Field(key="to_date", label="To"),
                dp.Field(key="status", label="Status"),
                dp.Field(key="occurrence", label="Occurrence"),
                dp.Field(key="changer", label="Changer"),
                dp.Field(key="changed", label="Changed"),
                dp.Field(key="type", label="Administrator"),
                dp.Field(key="review_task", label="Main Indicator Report"),
            ],
            legends=get_legends(ReviewComplianceTask),
            formatting=get_list_formatting(ReviewComplianceTask),
        )

    def get_instance_display(self) -> Display:
        nb_columns = 3
        fields = [
            ["title", "type", "review_task"],
            ["from_date", "to_date", "status"],
            ["is_instance", "occurrence", "occured"],
            [repeat_field(nb_columns, "description")],
        ]

        sections = []
        if self.view.kwargs.get("pk", None):
            instance = self.view.get_object()
            if not instance.is_instance:
                if instance.status == ReviewComplianceTask.Status.DRAFT:
                    fields.append([repeat_field(nb_columns, "task_group_section")])
                    sections.append(
                        create_simple_section(
                            "task_group_section", _("Groups Indicator"), [["task_group"]], "task_group", collapsed=True
                        )
                    )

                group_ids = instance.get_task_group_ids_from_review()
                no_group_ids = instance.get_task_group_ids_from_review(task_with_group=False)
                total = len(group_ids)
                if (
                    no_group_ids and instance.status != ReviewComplianceTask.Status.DRAFT
                ) or instance.status == ReviewComplianceTask.Status.DRAFT:
                    total += 1
                for count, group_id in enumerate(group_ids):
                    group = ComplianceTaskGroup.objects.get(id=group_id)
                    key = f"taskgroup{group.id}"
                    fields.append([repeat_field(nb_columns, f"{key}_section")])
                    sections.append(
                        create_simple_section(
                            f"{key}_section", f"({count + 1}/{total}). {group.name}", [[key]], key, collapsed=True
                        )
                    )

                if (
                    no_group_ids and instance.status != ReviewComplianceTask.Status.DRAFT
                ) or instance.status == ReviewComplianceTask.Status.DRAFT:
                    fields.append([repeat_field(nb_columns, "task_no_group_section")])
                    sections.append(
                        create_simple_section(
                            "task_no_group_section",
                            f'({total}/{total}). {_("No Indicator Group")}',
                            [["task_no_group"]],
                            "task_no_group",
                            collapsed=True,
                        )
                    )

            else:
                group_ids = instance.get_task_group_ids_from_review(through_task=False)
                no_group_ids = instance.get_task_group_ids_from_review(through_task=False, task_with_group=False)
                total = len(group_ids) + 2  # We add 2 to the total number to include Actions and events
                if no_group_ids:
                    total += 1
                for count, group_id in enumerate(group_ids):
                    group = ComplianceTaskGroup.objects.get(id=group_id)
                    key = f"taskinstancegroup{group.id}"
                    fields.append([repeat_field(nb_columns, f"{key}_section")])
                    sections.append(
                        create_simple_section(
                            f"{key}_section", f"({count + 1}/{total}). {group.name}", [[key]], key, collapsed=True
                        )
                    )

                if no_group_ids:
                    fields.append([repeat_field(nb_columns, "taskinstance_no_group_section")])
                    sections.append(
                        create_simple_section(
                            "taskinstance_no_group_section",
                            f'({count + 2}/{total}). {_("No Indicator Group")}',
                            [["taskinstance_no_group"]],
                            "taskinstance_no_group",
                            collapsed=True,
                        )
                    )

                fields.append([repeat_field(nb_columns, "actions_section")])
                sections.append(
                    create_simple_section(
                        "actions_section",
                        f'({total - 1}/{total}). {_("Actions")}',
                        [["actions"]],
                        "actions",
                        collapsed=True,
                    )
                )

                fields.append([repeat_field(nb_columns, "events_section")])
                sections.append(
                    create_simple_section(
                        "events_section", f'({total}/{total}). {_("Events")}', [["events"]], "events", collapsed=True
                    )
                )

        return create_simple_display(fields, sections)



import pytest
from wbcore.test import GenerateTest, default_config

config = {}
for key, value in default_config.items():
    config[key] = list(
        filter(
            lambda x: x.__module__.startswith("wbcompliance")
            and x.__name__
            not in [
                "RuleCheckedObjectRelationshipRiskRuleModelViewSet",
                "RuleThresholdRiskRuleModelViewSet",
                "ComplianceFormSignatureSectionRuleViewSet",
                "ComplianceFormSignatureModelViewSet",
            ],
            value,
        )
    )


@pytest.mark.django_db
@GenerateTest(config)
class TestProject:
    pass



from typing import Optional

from django.utils.translation import gettext as _
from wbcore.contrib.color.enums import WBColor
from wbcore.enums import Unit
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class RuleThresholdRiskRuleDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                # dp.Field(key="range", label=_("Range")),
                dp.Field(key="range_lower", label=_("Lower Range")),
                dp.Field(key="range_upper", label=_("Upper Range")),
                dp.Field(key="severity", label=_("Severity")),
                dp.Field(key="notifiable_users", label=_("Notifiable Users")),
                dp.Field(key="notifiable_groups", label=_("Notifiable Groups")),
                dp.Field(key="upgradable_after_days", label=_("Upgrade severity after")),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["severity", "upgradable_after_days", "range_lower", "range_upper"],
                [repeat_field(2, "notifiable_users"), repeat_field(2, "notifiable_groups")],
            ]
        )


class RuleCheckedObjectRelationshipRiskRuleDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                # dp.Field(key="checked_object_content_type",  label=_("Content Type")),
                # dp.Field(key="checked_object_id",  label=_("Object ID")),
                dp.Field(key="checked_object_repr", label=_("Checked Objects")),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display([["checked_object_id" if self.new_mode else "checked_object_repr"]])


class RiskRuleDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="name", label=_("Name"), width=Unit.PIXEL(300)),
                dp.Field(key="description", label=_("Description")),
                dp.Field(key="open_incidents_count", label=_("Open Incidents")),
                dp.Field(key="is_enable", label=_("Enable")),
                dp.Field(key="only_passive_check_allowed", label=_("Passive Only")),
                dp.Field(
                    key=None,
                    label="Administration",
                    open_by_default=False,
                    children=[
                        dp.Field(key="rule_backend", label=_("Rule Backend"), width=Unit.PIXEL(200)),
                        dp.Field(key="frequency", label=_("Frequency"), width=Unit.PIXEL(200)),
                        dp.Field(key="activation_date", label=_("Activation Date"), width=Unit.PIXEL(200)),
                        dp.Field(key="is_silent", label=_("Silent"), show="open"),
                        dp.Field(key="is_mandatory", label=_("Mandatory"), show="open"),
                        dp.Field(
                            key="automatically_close_incident", label=_("Automatically Close Incidents"), show="open"
                        ),
                        dp.Field(key="apply_to_all_active_relationships", label=_("All Checked Objects"), show="open"),
                        dp.Field(key="permission_type", label=_("Permission"), show="open"),
                    ],
                ),
            ],
            legends=[
                dp.Legend(
                    key="in_breach",
                    items=[
                        dp.LegendItem(icon=WBColor.RED_LIGHT.value, label=_("In Breach"), value="BREACH"),
                        dp.LegendItem(icon=WBColor.GREEN_LIGHT.value, label=_("Passed"), value="PASSED"),
                    ],
                ),
            ],
            formatting=[
                dp.Formatting(
                    column="in_breach",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.RED_LIGHT.value},
                            condition=("==", "BREACH"),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                            condition=("==", "PASSED"),
                        ),
                    ],
                )
            ],
        )

    def get_instance_display(self) -> Display:
        try:
            rule = self.view.get_object()
            parameter_fields = [
                [
                    f"parameters__{field}"
                    for field in rule.rule_backend.backend_class.get_serializer_class().get_parameter_fields()
                ]
            ]
        except Exception:
            parameter_fields = [["parameters"]]
        return create_simple_display(
            [
                [repeat_field(2, "name"), repeat_field(2, "rule_backend"), "is_enable"],
                ["permission_type", "frequency", "activation_date", repeat_field(2, "creator")],
                [
                    "only_passive_check_allowed",
                    "is_silent",
                    "is_mandatory",
                    "automatically_close_incident",
                    "apply_to_all_active_relationships",
                ],
                [repeat_field(5, "description")],
                [repeat_field(5, "parameters_section")],
                [repeat_field(5, "thresholds_section")],
                [repeat_field(5, "relationships_section")],
                [repeat_field(5, "incidents_section")],
            ],
            [
                create_simple_section("parameters_section", _("Parameters"), parameter_fields, collapsed=True),
                create_simple_section(
                    "thresholds_section", _("Thresholds"), [["thresholds"]], "thresholds", collapsed=True
                ),
                create_simple_section(
                    "relationships_section", _("Relationships"), [["relationships"]], "relationships", collapsed=True
                ),
                create_simple_section(
                    "incidents_section", _("Incidents"), [["incidents"]], "incidents", collapsed=False
                ),
            ],
        )



from django.contrib import admin

from wbcompliance.models.risk_management.incidents import (
    CheckedObjectIncidentRelationship,
    RiskIncident,
    RiskIncidentType,
)


class CheckedObjectIncidentRelationshipInLine(admin.TabularInline):
    model = CheckedObjectIncidentRelationship
    fields = ["status", "severity", "incident", "rule_check", "breached_value", "report_details", "report"]
    extra = 0
    raw_id_fields = ["incident", "rule_check"]
    readonly_fields = ["status", "severity", "incident", "rule_check"]


class IncidentInLine(admin.TabularInline):
    show_change_link = True
    model = RiskIncident
    fields = ["status", "severity", "date_range", "breached_content_type", "breached_object_id"]
    extra = 0


@admin.register(RiskIncidentType)
class RiskIncidentTypeModelAdmin(admin.ModelAdmin):
    list_display = [
        "name",
        "severity_order",
        "color",
        "is_ignorable",
        "is_automatically_closed",
        "is_informational",
    ]


@admin.register(RiskIncident)
class RiskIncidentModelAdmin(admin.ModelAdmin):
    list_display = [
        "status",
        "severity",
        "date_range",
        "rule",
        "breached_content_type",
        "breached_object_id",
        "resolved_by",
    ]

    inlines = [
        CheckedObjectIncidentRelationshipInLine,
    ]



from django.contrib import admin
from reversion_compare.admin import CompareVersionAdmin

from wbcompliance.models import (
    ComplianceForm,
    ComplianceFormRule,
    ComplianceFormSection,
    ComplianceFormSignature,
    ComplianceFormSignatureRule,
    ComplianceFormSignatureSection,
    ComplianceFormType,
)


@admin.register(ComplianceFormType)
class ComplianceFormTypeAdmin(CompareVersionAdmin):
    list_display = ["name", "type"]


@admin.register(ComplianceForm)
class ComplianceFormAdmin(CompareVersionAdmin):
    list_display = ["status", "creator", "created", "title"]

    autocomplete_fields = ["creator"]

    def reversion_register(self, model, **options):
        options = {
            "ignore_duplicates": True,
        }
        super().reversion_register(model, **options)


@admin.register(ComplianceFormSignature)
class ComplianceFormSignatureAdmin(CompareVersionAdmin):
    list_display = ["compliance_form", "version", "signed", "person"]

    autocomplete_fields = ["person"]


@admin.register(ComplianceFormSection)
class ComplianceFormSectionAdmin(CompareVersionAdmin):
    list_display = ["compliance_form", "name"]


@admin.register(ComplianceFormRule)
class ComplianceFormRuleAdmin(CompareVersionAdmin):
    list_display = ["section", "text", "ticked"]


@admin.register(ComplianceFormSignatureSection)
class ComplianceFormSignatureSectionAdmin(CompareVersionAdmin):
    list_display = ["compliance_form_signature", "name"]


@admin.register(ComplianceFormSignatureRule)
class ComplianceFormSignatureRuleAdmin(CompareVersionAdmin):
    list_display = ["section", "text", "ticked", "comments"]



from django.contrib import admin

from wbcompliance.models.risk_management.checks import RiskCheck


@admin.register(RiskCheck)
class RiskCheckModelAdmin(admin.ModelAdmin):
    list_display = ["rule", "creation_datetime", "evaluation_date"]
